#!/bin/bash
set -e
set -o pipefail

DB_FILE=.stumpf-prod.pgdump
BUCKET_NAME=diva-database-backups
REUSE_OLD_DB_DUMP=0
USE_RECENT_S3_DUMP=1

if ! [ -x "$(command -v aws)" ]; then
    echo 'Error: awscli is not installed.' >&2
    exit 1
fi

if ! [ -x "$(command -v jq)" ]; then
    echo 'Error: jq is not installed.' >&2
    exit 1
fi

while test $# -gt 0; do
    case "$1" in
        -h|--help)
            echo "import-production-db - Import a recent copy of the stumpf.kitware.com database locally"
            echo " "
            echo "import-production-db.sh [options]"
            echo " "
            echo "options:"
            echo "-h, --help                show brief help"
            echo "-r, --reuse               reuse an existing database dumped from this program"
            echo "-c, --current             import a copy of what the database looks like now, instead of using a recent copy"
            exit 0
            ;;
        -c|--current)
            USE_RECENT_S3_DUMP=0
            shift
            ;;
        -r|--reuse)
            REUSE_OLD_DB_DUMP=1
            shift
            ;;
        *)
            break
            ;;
    esac
done

# To customize postgres credentials for the local database
# use the various PG* environment variables.

if [ -z "$PGUSER" ]; then
    export PGUSER="postgres"
fi

if [ -z "$PGPASSWORD" ]; then
    export PGPASSWORD="postgres"
fi

if [ -z "$PGHOST" ]; then
    export PGHOST="localhost"
fi

if [ -z "$PGDATABASE" ]; then
    export PGDATABASE="stumpf"
fi

stumpf_secret () {
    aws secretsmanager get-secret-value \
        --secret-id "stumpf/rds_readonly.$1" | jq -r .SecretString
}

MOST_RECENT_S3_BACKUP=$(aws s3 ls "s3://$BUCKET_NAME" | sort | tail -n1 | awk '{ print $4; }')

if [[ "$REUSE_OLD_DB_DUMP" -eq 1 && -s "$DB_FILE" ]]; then
    echo "Skipping database dump, reusing $DB_FILE"
elif [[ "$USE_RECENT_S3_DUMP" -eq 1 && -n "$MOST_RECENT_S3_BACKUP" ]]; then
    echo "Pulling database dump from S3 [$MOST_RECENT_S3_BACKUP]"
    aws s3 cp "s3://$BUCKET_NAME/$MOST_RECENT_S3_BACKUP" "$DB_FILE"
else
    echo "Dumping production database"
    PGPASSWORD="$(stumpf_secret password)" pg_dump \
              --exclude-table-data "public.log" \
              -Fc --no-privileges \
              -U "$(stumpf_secret username)" \
              -h "$(stumpf_secret host)" stumpf > "$DB_FILE"
fi

dropdb --if-exists "$PGDATABASE"
createdb "$PGDATABASE"

echo "Restoring production database locally"
pg_restore -j4 --no-privileges --no-owner --dbname "$PGDATABASE" "$DB_FILE" > /dev/null
