import logging
import os
import time
from typing import Dict

from dotenv import load_dotenv
from flask import current_app, Flask, g, jsonify, request
from flask_cors import cross_origin
from flask_migrate import Migrate
from flask_sqlalchemy import get_debug_queries
from marshmallow import ValidationError
from pkg_resources import iter_entry_points
import sentry_sdk
from sentry_sdk.integrations.flask import FlaskIntegration
from sentry_sdk.integrations.logging import LoggingIntegration
from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration
import sqlalchemy as sa
from sqlalchemy import update
import statsd
from werkzeug.middleware.proxy_fix import ProxyFix

from stumpf.apispec import spec
from stumpf.auth import auth_bp, login_manager
from stumpf.decorators import public_endpoint
from stumpf.index import bp as index_bp
from stumpf.logging import ReloaderFilter
from stumpf.middleware import before_all_requests
from stumpf.mock_auth import mock_auth_bp
from stumpf.models import db, now_plus, Token, User
from stumpf.views import (annotation_app_bp, assignment_bp, batch_bp,
                          mturk_bp, qualification_type_bp, stats_bp,
                          task_bp, token_bp, user_bp, webhook_bp, worker_bp)


def handle_validation_error(e):
    return jsonify(e.messages), 400


def handle_general_error(e):
    current_app.logger.exception(e)
    return jsonify({'error': 'Something went wrong.'}), 500


def log_slow_db_queries(response):
    for query in get_debug_queries():
        if query.duration >= current_app.config['SLOW_QUERY_DURATION']:
            current_app.logger.info(f'Slow query took {query.duration:.2f}s: {query.statement}')

    return response


def initialize_development_user_token(app, db):
    with app.app_context():
        table_names = sa.inspect(db.engine).get_table_names()
        if not table_names and 'user' not in table_names:
            app.logger.error('Run stumpf-create-tables to initialize the database.')
            return

        user = User.query.filter_by(email='developer@kitware.com').first()

        try:
            if user:
                db.session.execute(update(Token.__table__)
                                   .where(Token.user_id == user.id)
                                   .values(expires=now_plus(days=365)()))
                db.session.commit()
                token = Token.query.filter_by(user=user).first()
            else:
                user = User(email='developer@kitware.com')
                token = Token(user=user, expires=now_plus(days=365)())
                db.session.add_all([user, token])
                db.session.commit()

            app.logger.info('Developer user created:')
            app.logger.info(f'Email: {user.email}')
            app.logger.info(f'Token: {token.token}')
        except Exception:
            app.logger.exception('Failed to create development user.')


def create_statsd(app: Flask):
    # Configuration is done from env vars, see:
    # https://statsd.readthedocs.io/en/v3.2.1/configure.html#from-the-environment
    # Also the client is fire-and-forget, so even if unconfigured it will silently
    # not track metrics.
    client = statsd.StatsClient(prefix='stumpf')

    @app.before_request
    def statsd_initialize():
        g.request_start = time.time()

    @app.after_request
    def statsd_submit_metrics(response):
        elapsed = time.time() - g.request_start

        if request.url_rule:  # some requests don't have a url_rule, particularly 404s
            for count in ['requests',
                          f'requests.{request.method.lower()}',
                          f'requests.{request.url_rule.rule}']:
                client.incr(count)
                client.timing(f'{count}.duration', elapsed)

        return response

    return client


def create_app(config: Dict) -> Flask:
    load_dotenv(os.getenv('DOTENV_PATH'))

    app = Flask(__name__)

    if app.config['ENV'] == 'development':
        app.logger.addFilter(ReloaderFilter())

    app.wsgi_app = ProxyFix(app.wsgi_app)
    app.config['SQLALCHEMY_RECORD_QUERIES'] = True
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    app.config['SQLALCHEMY_DATABASE_URI'] = os.getenv('SQLALCHEMY_DATABASE_URI')
    app.config['SLOW_QUERY_DURATION'] = float(os.getenv('SLOW_QUERY_DURATION', .75))
    app.config['BROKER_URL'] = os.getenv('BROKER_URL', 'localhost')
    app.config['SECRET_KEY'] = os.getenv('SECRET_KEY')
    app.config['STUMPF_PUBLIC_DOMAIN'] = os.getenv('STUMPF_PUBLIC_DOMAIN')
    app.config['GOOGLE_OAUTH_CLIENT_ID'] = os.getenv('GOOGLE_OAUTH_CLIENT_ID')
    app.config['GOOGLE_OAUTH_CLIENT_SECRET'] = os.getenv('GOOGLE_OAUTH_CLIENT_SECRET')
    app.config['WHITELISTED_EMAIL_ACCOUNTS'] = \
        [x for x in os.getenv('WHITELISTED_EMAIL_ACCOUNTS', '').split(',') if x]

    app.config['MTURK_SANDBOX_ENDPOINT_URL'] = os.getenv('MTURK_SANDBOX_ENDPOINT_URL')

    if app.config['ENV'] in ['development', 'testing']:
        app.logger.info(
            'Ignoring MTURK_PRODUCTION_ENDPOINT_URL, real HITs can only be launched in production.')
    else:
        app.config['MTURK_PRODUCTION_ENDPOINT_URL'] = os.getenv('MTURK_PRODUCTION_ENDPOINT_URL')

    app.config['MTURK_SNS_ARN'] = os.getenv('MTURK_SNS_ARN')

    if os.getenv('SENTRY_DSN', False):
        sentry_sdk.init(dsn=os.getenv('SENTRY_DSN'), environment=app.config['ENV'],
                        send_default_pii=True,
                        integrations=[FlaskIntegration(),
                                      LoggingIntegration(event_level=logging.WARNING),
                                      SqlalchemyIntegration()])

    app.config.update(config)

    db.init_app(app)
    Migrate(app, db)
    login_manager.init_app(app)

    # The default behavior in development mode will be to mock out google's
    # oauth service.  If a client secret key is set, the real oauth provider
    # will be used.
    if app.config['ENV'] == 'development' and \
            not app.config['GOOGLE_OAUTH_CLIENT_SECRET']:
        app.logger.info('Using mocked oauth service')
        app.register_blueprint(mock_auth_bp, url_prefix='/auth')
    else:
        app.logger.info('Using real oauth service')
        app.register_blueprint(auth_bp, url_prefix='/auth')

    app.before_request(before_all_requests)
    app.after_request(log_slow_db_queries)

    if app.config['ENV'] == 'development':
        initialize_development_user_token(app, db)

        if 'SQLALCHEMY_ECHO' in config and config['SQLALCHEMY_ECHO']:
            # Intentionally set this after creating the dev user/token
            app.config['SQLALCHEMY_ECHO'] = True

    app.register_blueprint(index_bp, url_prefix='/')
    app.register_blueprint(webhook_bp, url_prefix='/webhooks')

    app.register_blueprint(mturk_bp, url_prefix='/api/mturk/v1')

    app.register_blueprint(annotation_app_bp, url_prefix='/api/v1')
    app.register_blueprint(assignment_bp, url_prefix='/api/v1')
    app.register_blueprint(batch_bp, url_prefix='/api/v1')
    app.register_blueprint(qualification_type_bp, url_prefix='/api/v1')
    app.register_blueprint(stats_bp, url_prefix='/api/v1')
    app.register_blueprint(task_bp, url_prefix='/api/v1')
    app.register_blueprint(token_bp, url_prefix='/api/v1')
    app.register_blueprint(user_bp, url_prefix='/api/v1')
    app.register_blueprint(worker_bp, url_prefix='/api/v1')

    for ep in iter_entry_points('stumpf.pipeline.blueprints'):
        bp_loader = ep.load()
        bp_loader(app)

    # Only mask all errors/exceptions when in production
    if app.config['ENV'] == 'production':
        app.register_error_handler(500, handle_general_error)

    app.register_error_handler(ValidationError, handle_validation_error)

    # OpenAPI
    @app.route('/swagger.json', methods=['GET'])
    @public_endpoint
    @cross_origin()
    def swagger_json():
        return jsonify(spec.to_dict())

    app.statsd = create_statsd(app)

    with app.test_request_context():
        spec.path(view=app.view_functions['annotation_app.create_annotation_app'])
        spec.path(view=app.view_functions['annotation_app.list_annotation_app'])
        spec.path(view=app.view_functions['annotation_app.get_annotation_app'])
        spec.path(view=app.view_functions['annotation_app.delete_annotation_app'])

    return app
