from flask import abort, current_app, redirect, request, session
from flask_dance.consumer import oauth_authorized
from flask_dance.consumer.requests import OAuth2Session as BaseOAuth2Session
from flask_dance.consumer.storage.sqla import SQLAlchemyStorage
from flask_dance.contrib.google import make_google_blueprint
from flask_login import current_user, login_user
from flask_login import LoginManager
from sqlalchemy.orm.exc import NoResultFound

from stumpf.models import db, OAuth, Token, User


# This is a version of flask-dance's default session class, except that it
# sets the oauth state arg to a value from the flask session.
class OAuth2Session(BaseOAuth2Session):

    def authorization_url(self, url, **kwargs):
        kwargs.pop('state', None)  # ignore the state provided from flask-dance
        state = session.pop('oauth-redirect-destination', '/')
        return super().authorization_url(url, state=state, **kwargs)


auth_bp = make_google_blueprint(storage=SQLAlchemyStorage(OAuth, db.session, user=current_user),
                                reprompt_consent=True, offline=True, scope=['email'],
                                session_class=OAuth2Session)
login_manager = LoginManager()


@login_manager.user_loader
def load_user(user_id):
    return User.query.get(user_id)


def is_whitelisted_email(email):
    return email in current_app.config['WHITELISTED_EMAIL_ACCOUNTS']


# create/login local user on successful OAuth login
def google_logged_in(blueprint, token):
    if not token:
        abort(403)

    resp = blueprint.session.get('/oauth2/v2/userinfo')
    if not resp.ok:
        abort(403)

    google_info = resp.json()

    # hd is only guaranteed to exist for GSuite accounts, see
    # https://developers.google.com/identity/protocols/OpenIDConnect#hd-param
    if not is_whitelisted_email(google_info.get('email')) and \
            google_info.get('hd') != 'kitware.com':
        abort(403)

    oauth_query = OAuth.query.filter_by(
        provider=blueprint.name,
        provider_user_id=google_info['id']
    )

    try:
        oauth = oauth_query.one()
    except NoResultFound:
        oauth = OAuth(
            provider=blueprint.name,
            provider_user_id=google_info['id'],
            token=token
        )

    if oauth.user:
        login_user(oauth.user)
    else:
        try:
            user = User(email=google_info['email'])
            oauth.user = user
            db.session.add_all([user, oauth])
            db.session.commit()
            login_user(user)
        except Exception:
            db.session.rollback()
            raise

    stumpf_token = Token(user_id=current_user.get_id())
    db.session.add(stumpf_token)
    db.session.commit()

    response = redirect(request.args.get('state', '/'))
    response.set_cookie('stumpf-token',
                        stumpf_token.token,
                        domain=current_app.config.get('STUMPF_PUBLIC_DOMAIN'),
                        expires=stumpf_token.expires)
    return response


@oauth_authorized.connect_via(auth_bp)
def google_logged_in_decorated(*args, **kwargs):
    return google_logged_in(*args, **kwargs)
