from datetime import datetime, timedelta
from urllib.parse import urlencode

import boto3
import botocore
from celery.utils.log import get_task_logger
from flask import current_app
from sqlalchemy import func, or_

from stumpf.celery.app import app
from stumpf.models import (Assignment, AssignmentStatusChange, db, KwitLease, MturkToken,
                           Task, TaskStatusChange, Token)
from stumpf.mturk_utils import create_hit, question_xml

logger = get_task_logger(__name__)


@app.task
def abandon_expired_kwit_assignments():
    expiration_time = Assignment.created + func.make_interval(0, 0, 0, 0, 0, 0,
                                                              Task.assignment_duration_in_seconds)
    one_week_ago = datetime.utcnow() - timedelta(weeks=1)
    expired_assignments = db.session \
        .query(Task, Assignment) \
        .filter(Task.id == Assignment.task_id) \
        .filter(Task.type_ == 'KWIT') \
        .filter(Assignment.status == 'Accepted') \
        .filter(or_(expiration_time < datetime.utcnow(), Assignment.created < one_week_ago)) \
        .all()

    for task, assignment in expired_assignments:
        current_app.logger.info(f'Expiring {task.annotation_app.name}/{assignment.id}')
        status_change = AssignmentStatusChange(assignment_id=assignment.id,
                                               status='Abandoned')
        task_status_change = TaskStatusChange(task=task, status='Assignable')
        KwitLease.query.filter_by(assignment=assignment).delete()
        db.session.add_all([status_change, task_status_change])
        db.session.commit()


@app.task
def clean_expired_tokens():
    Token.query.filter(Token.expired).delete()
    MturkToken.query.filter(MturkToken.expired).delete()
    db.session.commit()


@app.task
def test_flask_context():
    return current_app.config['ENV']


@app.task
def test_db_connection():
    result = db.session.execute('SELECT COUNT(*) FROM assignment;')
    return result.scalar()


@app.task(bind=True, rate_limit='90/m', acks_late=True)
def create_hit_task(self, user_id, app_id, app_url, hit_args):
    # Note: We do not do any validation of arguments inside the task
    # to prevent swamping the database with checks to see if
    # individual video_ids etc exists. This validation is handled
    # en-mass in the endpoint that calls the task.
    session = boto3.session.Session()
    if hit_args['sandbox']:
        mturk = session.client('mturk',
                               endpoint_url=current_app.config['MTURK_SANDBOX_ENDPOINT_URL'])
    else:
        mturk = session.client('mturk',
                               endpoint_url=current_app.config['MTURK_PRODUCTION_ENDPOINT_URL'])

    # TODO: possible to make hit creation idempotent by querying mturk?
    try:
        task = Task(annotation_app_id=app_id,
                    user_id=user_id,
                    args=hit_args.get('args', {}),
                    assignment_duration_in_seconds=hit_args['assignment_duration_in_seconds'],
                    description=hit_args['description'],
                    keywords=hit_args.get('keywords', ''),
                    max_assignments=hit_args['max_assignments'],
                    qualification_for_id=hit_args.get('qualification_for'),
                    qualification_requirements=hit_args['qualification_requirements'],
                    reward=hit_args['reward'],
                    sandbox=hit_args['sandbox'],
                    title=hit_args['title'],
                    type_='HIT')

        status_change = TaskStatusChange(task=task, status='Assignable')
        db.session.add_all([task, status_change])
        db.session.commit()
        try:
            task_id = task.id

            hit = create_hit(mturk, hit_args['assignment_duration_in_seconds'],
                             hit_args['reward'], hit_args['title'], hit_args['description'],
                             hit_args.get('keywords', ''), hit_args['max_assignments'],
                             hit_args['lifetime_in_seconds'],
                             question_xml(app_url=app_url,
                                          query_string=urlencode({**hit_args.get('args', {})})),
                             hit_args['auto_approval_delay_in_seconds'],
                             hit_args['qualification_requirements'])

            task.hit_id = hit['HIT']['HITId']
            task.hit_group_id = hit['HIT']['HITGroupId']
            db.session.add(task)
            db.session.commit()

            logger.info(f'Created HIT {hit["HIT"]["HITId"]} for Task {task_id}')

        except botocore.exceptions.ClientError as exc:
            task_id = task.id
            # Remove the task/status change and retry the task after 5 seconds
            db.session.delete(status_change)
            db.session.delete(task)
            db.session.commit()

            logger.warning(f'Boto client error when creating HIT, removed task: {task_id}')

            max_retries = 3

            if (self.request.retries + 1) > max_retries:
                # TODO: notify this information via some mechanism (sentry? slack?)
                logger.error(f'Exceeded max retries, task {task_id} failed!')

            self.retry(exc=exc, countdown=5, max_retries=max_retries)


#        TODO: How do we indicate to statsd that we have created an
#         active HIT now that this is being handled by a remote
#         worker?
#        current_app.statsd.incr('hits.active')

    except Exception:
        db.session.rollback()
        raise

    return hit
