from celery import Celery
from celery.signals import task_postrun, worker_process_init
import pkg_resources as pr

from stumpf import cached_iter_entry_points

PIPELINE_ENTRYPOINTS = [
    'stumpf.pipeline.tasks.assignment_status_change',
    'stumpf.pipeline.tasks.task_status_change'
]


def push_flask_context_on_worker_start(*args, **kwargs):
    from stumpf.app import create_app
    flask_app = create_app({})
    flask_app.app_context().push()


def rollback_database_on_task_end(*args, **kwargs):
    from stumpf.models import db
    db.session.rollback()


def make_celery(entrypoint_config=True):
    # Make sure the flask app context is pushed on celery init
    # This is similar to decorating with @celeryd_init.connect
    # Except we do it as apart of the celery application factory
    # rather then as the function is being defined.
    worker_process_init.connect()(push_flask_context_on_worker_start)
    task_postrun.connect()(rollback_database_on_task_end)

    # Create the celery application object
    celery_app = Celery()

    # Pull in additional configuration from CELERY_CONFIG_MODULE,
    # default is stumpf.celery.config.
    if entrypoint_config:
        # Always load core config first
        for ep in pr.iter_entry_points('stumpf.pipeline.celery.conf', 'core'):
            func = ep.load()
            func(celery_app)

        for ep in pr.iter_entry_points('stumpf.pipeline.celery.conf'):
            if ep.name != 'core':
                func = ep.load()
                func(celery_app)

    for pipeline_entrypoint in PIPELINE_ENTRYPOINTS:
        for ep in cached_iter_entry_points(pipeline_entrypoint):
            if ep.module_name not in celery_app.conf.imports:
                celery_app.conf.imports += (ep.module_name, )

    return celery_app
