import random
from uuid import UUID

import click
from dotenv import find_dotenv, load_dotenv
import flask_migrate
from tabulate import tabulate

from stumpf.app import create_app
from stumpf.cli.trace import hit_history, print_task, table_display_from_history
from stumpf.fakes import fake
from stumpf.models import db, Task, User


load_dotenv(find_dotenv(usecwd=True))


@click.command()
@click.argument('what')
def trace(what):
    """
    Trace the lifecycle of a Task through the system.

    WHAT can be a Stumpf task identifier or a Mechanical Turk HIT ID.
    """
    try:
        UUID(what)
        is_stumpf_task = True
    except ValueError:
        is_stumpf_task = False

    app = create_app({
        'SQLALCHEMY_ECHO': False
    })
    with app.app_context():
        if is_stumpf_task:
            task = Task.query.get(what)
            hit_id = task.hit_id
        else:
            task = Task.query.filter_by(hit_id=what).first()
            hit_id = what

        history = hit_history(hit_id)
        table = table_display_from_history(history)

        if not task and not table:
            click.secho(f'Nothing found for ID {what}', fg='red')
        elif not task:
            click.secho('No corresponding Stumpf task found, was it created manually?', fg='yellow')
            click.echo('')
            click.echo('History')
            click.echo(tabulate(table))
        elif not table:
            click.secho('No history found for this task, does it exist?', fg='yellow')
            click.echo('')
            click.echo('Task')
            print_task(task)
        else:
            click.echo('Task')
            print_task(task)
            click.echo('')
            click.echo('History')
            click.echo(tabulate(table))


@click.command()
@click.option('--only-if-empty', type=click.BOOL, default=False, is_flag=True,
              help='Only generate data if the DB is empty (defaults to false).')
@click.option('--small', type=click.BOOL, default=False, is_flag=True,
              help='Generate a minimal amount of data for testing.')
def sample_data(only_if_empty, small):
    """
    Management script for sample data generation.
    """
    app = create_app({})

    grouper_app = {
        'repo_url': 'https://github.com/Kitware/vaui',
        'name': 'Grouper',
        'url': 'https://mturk.kitware.com/grouper'}
    refiner_app = {
        'repo_url': 'https://github.com/Kitware/refiner',
        'name': 'Refiner',
        'url': 'https://mturk.kitware.com/refiner'}

    with app.app_context():
        if only_if_empty and User.query.first() is not None:
            return

        fake_users = [fake.stumpf_user() for _ in range(25)]
        db.session.add_all(fake_users)
        db.session.commit()

        annotation_apps = [
            fake.stumpf_annotation_app(**grouper_app),
            fake.stumpf_annotation_app(**refiner_app)]
        annotation_apps.extend([fake.stumpf_annotation_app() for _ in range(4)])

        try:
            for ann_app in annotation_apps:
                db.session.add(ann_app)
                db.session.commit()
                for _ in range(10 if small else 1000):
                    for _ in range(10):
                        task = fake.stumpf_task(ann_app.id, random.choice(fake_users).id)
                        db.session.add(task)
                        db.session.commit()

                        for _ in range(3):
                            assignment = fake.stumpf_assignment(task.id)
                            db.session.add(assignment)
                            db.session.commit()

                            changes = fake.stumpf_assignment_status_changes(assignment)
                            db.session.add_all(changes)
                            db.session.commit()

                db.session.commit()

        except Exception:
            click.echo('Failed to generate random data', err=True)
            raise


@click.command()
def create_tables():
    app = create_app({})

    with app.app_context():
        db.create_all()
        flask_migrate.stamp()
