import json

import click
from dateutil import parser
import pytz
from tabulate import tabulate

from stumpf.models import Log, Task


status_colors = {
    'Assignable': 'green',
    'Unassignable': 'green',
    'Reviewable': 'yellow',
    'Reviewing': 'yellow',
    'Disposed': 'red',
    'Expired': 'red',
}


def utcify(dt):
    return dt.replace(tzinfo=pytz.timezone('UTC'))


def print_task(task):
    table = []
    table.append(['id', task.id])
    table.append(['hit_id', task.hit_id])
    table.append(['created', task.created])
    table.append(['type_', task.type_])
    table.append(['title', task.title])
    table.append(['app', task.annotation_app.name])
    table.append(['max_assignments', task.max_assignments])
    table.append(['status', click.style(task.status, fg=status_colors[task.status])])
    table.append(['args', json.dumps(task.args, indent=2)])
    click.echo(tabulate(table))


def sns_history(hit_id):
    history = []

    sns_messages = Log.query.filter_by(category='raw-sns-message')\
        .filter(Log.message[('Message', 'Events', '0', 'HITId')].astext == hit_id).all()

    for sns_message in sns_messages:
        dt = utcify(parser.parse(sns_message.message['Message']['Events'][0]['EventTimestamp']))
        history.append(('SNS Message', dt, sns_message))

    return sorted(history, key=lambda x: x[1])


def hit_history(hit_id):
    history = []

    maybe_task = Task.query.filter_by(hit_id=hit_id).first()

    if maybe_task:
        history.append(('Task Created', utcify(maybe_task.created), maybe_task))

        for status_change in maybe_task.status_changes:
            history.append(('Status Changed', utcify(status_change.created), status_change))

    history += sns_history(hit_id)

    return sorted(history, key=lambda x: x[1])


def table_display_from_history(history):
    def format_time(t):
        return t.strftime('%e %b %Y %H:%M:%S%p')

    table = []

    for (event_type, event_time, data) in history:
        if event_type == 'Task Created':
            table.append(['Task Created', format_time(data.created)])
        elif event_type == 'Status Changed':
            status_str = 'None'
            if data.previous_status_occurred:
                status_str = click.style(data.previous_status_occurred.status,
                                         fg=status_colors[data.previous_status_occurred.status])

            new_status = click.style(data.status, fg=status_colors[data.status])
            s = f'{status_str} -> {new_status}'
            table.append(['Status Changed', format_time(data.created), s])
        else:
            event = data.message['Message']['Events'][0]
            if event['EventType'].startswith('Assignment'):
                continue

            table.append(['SNS Message', format_time(data.created), event['EventType']])

    return table
