from collections import OrderedDict
from datetime import datetime
import json
import secrets
import string

from faker import Faker
from faker.providers import BaseProvider

from stumpf.models import AnnotationApp, Assignment, AssignmentStatusChange, QualificationType, \
    Task, User, Worker

fake = Faker()
alphabet = string.ascii_uppercase + string.digits


class MTurkProvider(BaseProvider):

    def hit_id(self):
        pass

    def assignment_id(self):
        return ''.join(secrets.choice(alphabet) for i in range(30))

    def sns_message(self, events=None):
        return {
            'Type': 'Notification',
            'MessageId': self.generator.uuid4(),
            'TopicArn': 'arn:aws:sns:us-east-2:066423736832:mturk-notifications',
            'Subject': '18151108',
            'Timestamp': '2018-06-19T12:57:07.894Z',
            'SignatureVersion': '1',
            'Signature': 'fake-signature',
            'SigningCertURL': 'https://sns.us-east-2.amazonaws.com/fake-sign-cert-url',
            'UnsubscribeURL': 'https://sns.us-east-2.amazonaws.com/fake-unsub-url',
            'Message': json.dumps({
                'Events': events if events is not None else []
            })
        }

    def sns_task_event(self, **kwargs):
        event = {
            'HITGroupId': ''.join(secrets.choice(alphabet) for i in range(30)),
            'EventType': self.generator.random_element(
                ['HITCreated', 'HITExpired', 'HITReviewable', 'HITDisposed']),
            'EventTimestamp': datetime.utcnow().isoformat(timespec='seconds') + 'Z',
            'HITId': ''.join(secrets.choice(alphabet) for i in range(30)),
            'HITTypeId': ''.join(secrets.choice(alphabet) for i in range(30)),
        }

        event.update(kwargs)

        return event

    def sns_assignment_event(self, **kwargs):
        event = {
            'HITGroupId': ''.join(secrets.choice(alphabet) for i in range(30)),
            'EventType': self.generator.random_element(
                ['AssignmentAbandoned', 'AssignmentReturned', 'AssignmentAccepted',
                 'AssignmentSubmitted', 'AssignmentApproved', 'AssignmentRejected']),
            'EventTimestamp': datetime.utcnow().isoformat(timespec='seconds') + 'Z',
            'HITId': ''.join(secrets.choice(alphabet) for i in range(30)),
            'AssignmentId': ''.join(secrets.choice(alphabet) for i in range(30)),
            'WorkerId': ''.join(secrets.choice(alphabet) for i in range(14)),
            'HITTypeId': ''.join(secrets.choice(alphabet) for i in range(30)),
            'EventDocId': '065146d07bd633ed9dc745c3ed6894cbf44c6182',
            'SourceAccount': '066423736832',
            'CustomerId': 'A2TCFET4XJ6CM1',
            'EventDocVersion': '2014-08-15'
        }

        event.update(kwargs)

        return event


class StumpfProvider(BaseProvider):
    def stumpf_user(self):
        return User(email=self.generator.email(domain='kitware.com'))

    def stumpf_task(self, annotation_app_id, user_id, status=None):
        return Task(**{
            'annotation_app_id': annotation_app_id,
            'user_id': user_id,
            'sandbox': self.generator.boolean(),
            'hit_id': self.generator.uuid4(),
            'type_': self.generator.random_element(OrderedDict(
                [('HIT', 0.95), ('KWIT', 0.05)]
            )),
            'title': f'{self.generator.word()} Video annotation',
            'description': self.generator.paragraph(),
            'keywords': ','.join(self.generator.words(nb=5)),
            'assignment_duration_in_seconds': self.generator.random_int(min=60, max=900),
            'status': status if status else
                        self.generator.random_element(['Assignable', 'Unassignable', 'Reviewable',
                                                       'Reviewing', 'Disposed']),
            'review_status': self.generator.random_element(
                ['NotReviewed', 'MarkedForReview', 'ReviewedAppropriate', 'ReviewedInappropriate']),
            'max_assignments': self.generator.random_int(min=1, max=5),
            'reward': self.generator.random_int(min=10, max=200),
            'args': self.generator.pydict(3, False, 'str')
        })

    def stumpf_assignment(self, task_id, app_version=None):
        task = Task.query.get(task_id)
        assignment_kwargs = {}

        if task.type_ == 'HIT':
            assignment_kwargs.update({
                'assignment_id': ''.join(secrets.choice(alphabet) for i in range(30)),
                'worker_id': ''.join(secrets.choice(alphabet) for i in range(14)),
                'app_version': app_version if app_version is not None else self.generator.sha1(),
                'status': 'Accepted'
            })
        elif task.type_ == 'KWIT':
            assignment_kwargs.update({
                'app_version': app_version if app_version is not None else self.generator.sha1(),
                'status': 'Accepted'
            })

        assignment = Assignment(task_id=task_id, **assignment_kwargs)
        return assignment

    def stumpf_assignment_status_changes(self, assignment):
        assignment_status = self.generator.random_element(
            ['Abandoned', 'Returned', 'Accepted', 'Submitted', 'Approved', 'Rejected'])
        changes = [(AssignmentStatusChange(occurred=datetime.utcnow(),
                                           assignment_id=assignment.id,
                                           status='Accepted'))]

        if assignment_status == 'Accepted':
            return changes
        elif assignment_status in ('Returned', 'Abandoned'):
            changes.append(AssignmentStatusChange(occurred=datetime.utcnow(),
                                                  assignment_id=assignment.id,
                                                  status=assignment_status))
            return changes
        else:
            changes.append(AssignmentStatusChange(occurred=datetime.utcnow(),
                                                  assignment_id=assignment.id,
                                                  status='Submitted'))

        if assignment_status == 'Submitted':
            return changes
        elif assignment_status in ('Approved', 'Rejected'):
            changes.append(AssignmentStatusChange(occurred=datetime.utcnow(),
                                                  assignment_id=assignment.id,
                                                  status=assignment_status))

        return changes

    def stumpf_annotation_app(self, repo_url=None, hash_=None, name=None, url=None):
        return AnnotationApp(**{
            'repo_url': repo_url if repo_url is not None else self.generator.url(),
            'name': name if name is not None else f'{self.generator.last_name()} App',
            'url': url if url is not None else self.generator.url()
        })

    def stumpf_worker(self):
        return Worker(
            mturk_id=''.join(secrets.choice(alphabet) for i in range(14)),
            meta={}
        )

    def stumpf_qualification_type(self, user_id, mturk_qt_id=None):
        if mturk_qt_id is None:
            mturk_qt_id = ''.join(secrets.choice(alphabet) for i in range(20))

        return QualificationType(**{
            'user_id': user_id,
            'mturk_qt_id': mturk_qt_id,
            'sandbox': self.generator.boolean(),
            'name': self.generator.text(max_nb_chars=32),
            'description': self.generator.sentence(),
            'keywords': ' '.join(self.generator.words(nb=2)),
            'status': 'Active'
        })


fake.add_provider(StumpfProvider)
fake.add_provider(MTurkProvider)
