from flask import request
import sqlalchemy

from stumpf.sorting import boolean


def filter_query_string(model, field, query=None):
    """
    Return a filtered query from the provided model based on a string field.

    :param model: The sqlachemy model definition
    :param query: The query to sort (defaults to model.query)
    """
    search_value = request.args.get(field, type=str)

    query = query or model.query

    if search_value is not None:
        query = query.filter(model.searchable_fields[field].contains(search_value))

    return query


def filter_multifield_query(model, fields, querykey, query=None):
    """
    Returns a filtered query from the provided model

    :param model: the sqlalchemy model definition
    :param fields: a list of strings representing fields on the model to search
    :param querykey: the key in the request that contains the search terms
    """
    search_value = request.args.get(querykey, type=str)

    query = query or model.query

    if search_value is not None:
        filter_args = tuple(sqlalchemy.cast(
            model.searchable_fields[f], sqlalchemy.String).contains(search_value) for f in fields)
        query = query.filter(sqlalchemy.or_(*filter_args))

    return query


def filter_query_int(model, field, query=None):
    """
    Return a filtered query from the provided model based on an int field.

    :param model: The sqlachemy model definition
    :param query: The query to sort (defaults to model.query)
    """
    search_value = request.args.get(field, type=int)

    query = query or model.query

    if search_value is not None:
        query = query.filter(model.searchable_fields[field] == search_value)

    return query


def filter_query_enum(model, field, query=None):
    """
    Return a filtered query from the provided model based on an enum field.

    :param model: The sqlachemy model definition
    :param query: The query to sort (defaults to model.query)
    """
    search_value = request.args.get(field, type=str)

    query = query or model.query

    if search_value is not None and search_value != '':
        query = query.filter(model.searchable_fields[field] == search_value)

    return query


def filter_query_bool(model, field, query=None):
    """
    Return a filtered query from the provided model based on a boolean field.

    :param model: The sqlachemy model definition
    :param query: The query to sort (defaults to model.query)
    """
    search_value = request.args.get(field, type=str)

    query = query or model.query

    if search_value is not None:
        query = query.filter(model.searchable_fields[field] == boolean(search_value))

    return query
