from datetime import datetime, timedelta
import secrets
from typing import Optional
import uuid

from flask_dance.consumer.storage.sqla import OAuthConsumerMixin
from flask_sqlalchemy import SQLAlchemy
from marshmallow import fields, post_load, Schema, validate, validates, ValidationError
from psycopg2.extras import NumericRange
from sqlalchemy import DDL, event, MetaData
from sqlalchemy.dialects import postgresql as pg
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.schema import UniqueConstraint

from stumpf import cached_iter_entry_points
from stumpf.apispec import spec


class BaseSQLAlchemy(SQLAlchemy):
    def create_all(self, *args, **kwargs):
        self.import_pipeline_models()
        super().create_all(*args, **kwargs)

    def import_pipeline_models(self):
        for ep in cached_iter_entry_points('stumpf.pipeline.models'):
            # Crash if this fails to load
            ep.load()


# This is to avoid having to manually name all constraints
# See: http://alembic.zzzcomputing.com/en/latest/naming.html
metadata = MetaData(naming_convention={
    'pk': 'pk_%(table_name)s',
    'fk': 'fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s',
    'ix': 'ix_%(table_name)s_%(column_0_name)s',
    'uq': 'uq_%(table_name)s_%(column_0_name)s',
    'ck': 'ck_%(table_name)s_%(constraint_name)s',
})
db = BaseSQLAlchemy(metadata=metadata)
event.listen(db.Model.metadata, 'before_create', DDL('CREATE EXTENSION IF NOT EXISTS "uuid-ossp";'))
event.listen(db.Model.metadata, 'before_create', DDL('CREATE EXTENSION IF NOT EXISTS "tablefunc";'))


class BaseSchema(Schema):
    __model__ = None

    id = fields.UUID()
    created = fields.DateTime(dump_only=True)

    @post_load
    def make_object(self, data, **kwargs):
        return self.__model__(**data)


class Range(fields.List):
    """
    This serializes a numeric interval as a pair of numbers, where the first
    number is the minimum value *inclusive* and the second number is the
    maximum value *exclusive*.  The serialization assumes that the numeric
    ranges are integers.

    For example, the closed interval [1, 100] serializes as follows:

        NumericRange(1, 100, bounds='[]') -> [1, 101]
    """
    @classmethod
    def numeric_range_to_list(cls, value):
        # handle inclusive/exclusive range values correctly
        lower = value.lower
        if not value.lower_inc:
            lower += 1
        upper = value.upper
        if value.upper_inc:
            upper += 1
        return [lower, upper]

    def _deserialize(self, value, attr, data):
        a, b = super()._deserialize(value, attr, data)
        return NumericRange(a, b - 1, bounds='[]')

    def _serialize(self, value, attr, obj):
        return super()._serialize(
            self.numeric_range_to_list(value), attr, obj)


class User(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    email = db.Column(db.String, unique=True, nullable=False)

    @property
    def is_authenticated(self):
        return True

    @property
    def is_active(self):
        return True

    @property
    def is_anonymous(self):
        return not self.is_authenticated

    def get_id(self):
        return self.id


class UserSchema(BaseSchema):
    __model__ = User

    email = fields.Email(required=True)


class OAuth(OAuthConsumerMixin, db.Model):
    provider_user_id = db.Column(db.String(256), unique=True)
    user_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey(User.id))
    user = db.relationship(User)


def now_plus(**kwargs):
    def f():
        return datetime.utcnow() + timedelta(**kwargs)
    return f


class Token(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    user_id = db.Column(db.ForeignKey(User.id, name='token_user_id_fkey'), nullable=False)
    user = db.relationship(User)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    expires = db.Column(db.DateTime, nullable=False, default=now_plus(days=7))
    token = db.Column(db.String, index=True, unique=True, nullable=False, default=secrets.token_hex)

    @classmethod
    def user_from_token(cls, token: str) -> Optional[User]:
        t = cls.query.filter_by(token=token).filter(cls.expires > datetime.utcnow()).first()

        if t:
            return t.user

    @hybrid_property
    def expired(self):
        return self.expires < datetime.utcnow()


event.listen(Token.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE VIEW view_token AS (
                SELECT id, user_id, created, expires
                FROM token)
             """))


class TokenSchema(BaseSchema):
    __model__ = Token

    user_id = fields.UUID(load_only=True)
    user = fields.Nested('UserSchema', dump_only=True)
    expires = fields.DateTime()
    token = fields.Str(dump_only=True)


class MturkToken(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    assignment_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('assignment.id'), nullable=False)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    expires = db.Column(db.DateTime, nullable=False)
    token = db.Column(db.String, unique=True, nullable=False, default=secrets.token_hex)

    assignment = db.relationship('Assignment', backref=db.backref('mturk_tokens', lazy=True))

    @hybrid_property
    def expired(self):
        return self.expires < datetime.utcnow()


event.listen(MturkToken.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE VIEW view_mturk_token AS (
                SELECT id, assignment_id, created, expires
                FROM mturk_token)
             """))


class MturkTokenSchema(BaseSchema):
    __model__ = MturkToken

    expires = fields.DateTime()
    token = fields.Str()
    assignment = fields.Nested('AssignmentSchema', dump_only=True)


class AnnotationApp(db.Model):
    __table_args__ = (UniqueConstraint('repo_url', 'name', 'url'), )

    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    repo_url = db.Column(db.String, nullable=False)
    name = db.Column(db.String, nullable=False)
    url = db.Column(db.String, nullable=False)

    sortable_fields = {
        'created': created,
        'name': name
    }
    default_sort = 'created'
    default_desc = True


class AnnotationAppSchema(BaseSchema):
    __model__ = AnnotationApp

    repo_url = fields.Url(required=True)
    url = fields.Url(required=True)
    name = fields.Str(required=True)


spec.components.schema('AnnotationApp', schema=AnnotationAppSchema)
spec.tag({'name': 'AnnotationApp'})


class Task(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    user_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('user.id'), nullable=False)
    user = db.relationship('User', backref=db.backref('tasks', lazy=True))
    annotation_app_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('annotation_app.id'),
                                  nullable=False)
    annotation_app = db.relationship('AnnotationApp', backref=db.backref('tasks', lazy=True))
    sandbox = db.Column(pg.BOOLEAN, nullable=False, default=False)
    hit_id = db.Column(db.String, nullable=True, unique=True)
    hit_group_id = db.Column(db.String, nullable=True)
    type_ = db.Column(pg.ENUM('HIT', 'KWIT', name='hit_kwit_enum'), nullable=False)
    title = db.Column(db.String, nullable=False)
    description = db.Column(db.String, nullable=True)
    keywords = db.Column(db.String, nullable=True)
    assignment_duration_in_seconds = db.Column(db.Integer, nullable=True)
    status = db.Column(pg.ENUM('Assignable', 'Unassignable', 'Reviewable', 'Reviewing', 'Disposed',
                               'Expired',
                               name='task_status_enum'), nullable=False, default='Assignable')
    review_status = db.Column(pg.ENUM('NotReviewed', 'MarkedForReview', 'ReviewedAppropriate',
                                      'ReviewedInappropriate', name='task_review_status_enum'),
                              nullable=False, default='NotReviewed')
    max_assignments = db.Column(db.Integer, nullable=False)
    reward = db.Column(db.Integer, nullable=True)  # cents
    args = db.Column(pg.JSONB, nullable=False, default=lambda: {})
    answer = db.Column(pg.JSONB, nullable=True)
    qualification_requirements = db.Column(pg.JSONB, nullable=False, default=lambda: [])
    qualification_for_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('qualification_type.id'),
                                     nullable=True)
    qualification_for = db.relationship('QualificationType', backref=db.backref('tasks', lazy=True))

    searchable_fields = {
        'type': type_,
        'task_name': title,
        'assignments': max_assignments,
        'status': status,
        'sandbox': sandbox
    }

    sortable_fields = {
        'created': created,
        'title': title,
        'type_': type_
    }
    default_sort = 'created'
    default_desc = True

    @property
    def status_changes(self):
        return TaskStatusChange.query.filter_by(task_id=self.id) \
                                     .order_by('occurred').all()


class TaskSchema(BaseSchema):
    __model__ = Task

    annotation_app = fields.Nested('AnnotationAppSchema', dump_only=True)
    title = fields.Str(required=True)
    type_ = fields.Str(required=True)
    sandbox = fields.Boolean(required=True)
    hit_id = fields.Str()
    hit_group_id = fields.Str()
    title = fields.Str(required=True)
    description = fields.Str(required=True)
    keywords = fields.Str()
    assignment_duration_in_seconds = fields.Int()
    status = fields.Str()
    review_status = fields.Str()
    max_assignments = fields.Int(required=True, validate=validate.Range(min=1, max=10))
    reward = fields.Int(required=True, validate=validate.Range(min=1, max=1000))
    args = fields.Dict()
    answer = fields.Dict()
    qualification_for_id = fields.UUID(required=False, load_only=True)
    qualification_for = fields.Nested('QualificationTypeSchema', dump_only=True)
    qualification_requirements = fields.List(fields.Dict())


class TaskStatusChange(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    occurred = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    task_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey(Task.id), nullable=False)
    status = db.Column(pg.ENUM('Assignable', 'Unassignable', 'Reviewable', 'Reviewing', 'Disposed',
                               'Expired',
                               name='task_status_enum'), nullable=False)
    task = db.relationship('Task')

    @property
    def previous_status_occurred(self):
        return TaskStatusChange.query.filter_by(task_id=self.task_id) \
                                     .filter(TaskStatusChange.occurred < self.occurred) \
                                     .order_by(TaskStatusChange.occurred.desc()).first()


@event.listens_for(TaskStatusChange, 'after_insert')
def task_status_change_hook(mapper, connection, task_status_change):
    from flask import current_app as flask_app
    if task_status_change.status in ['Reviewable', 'Expired']:
        celery_task_ids = []
        for ep in cached_iter_entry_points('stumpf.pipeline.tasks.task_status_change'):
            try:
                flask_app.logger.info(
                    f'Calling task_status_change hook: {ep.name} on {task_status_change.task_id}')
                func = ep.load()
                # TODO: Better error checking here
                async_res = func.delay(task_status_change.id,
                                       task_status_change.task_id,
                                       task_status_change.status)
                celery_task_ids.append(str(async_res.id))
            except ImportError:
                flask_app.logger.error(f'Failed to load task_status_change hook: {ep.name}')

        if len(celery_task_ids):
            connection.execute(insert(Log.__table__).values(
                category='task-status-change-event',
                message={
                    'status_change': str(task_status_change.id),
                    'celery_ids': celery_task_ids}))
        else:
            flask_app.logger.error('Expected at least one task_status_change handler')


event.listen(TaskStatusChange.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE FUNCTION update_task_status()
             RETURNS TRIGGER AS
             $BODY$
             BEGIN
                UPDATE task SET status = (
                    SELECT status FROM task_status_change
                    WHERE task_id = NEW.task_id
                    ORDER BY occurred DESC LIMIT 1)
                WHERE id = NEW.task_id;
             RETURN NEW;
             END;
             $BODY$
             LANGUAGE plpgsql VOLATILE;"""))
event.listen(TaskStatusChange.__table__, 'after_create',
             DDL("""
             CREATE TRIGGER update_task_status
             AFTER INSERT OR UPDATE ON task_status_change
             FOR EACH ROW EXECUTE PROCEDURE update_task_status();"""))


class CreateTaskSchema(Schema):
    sandbox = fields.Bool(missing=True)
    title = fields.Str(required=True)
    description = fields.Str(required=True)
    keywords = fields.Str()
    max_assignments = fields.Int(required=True, validate=validate.Range(min=1, max=1000))
    args = fields.Dict(keys=fields.Str())


class CreateKwitSchema(CreateTaskSchema):
    @validates('max_assignments')
    def validate_max_assignments(self, max_assignments):
        if max_assignments != 1:
            raise ValidationError('Max assignments must be exactly 1 for KWITs.')


class CreateHitSchema(CreateTaskSchema):
    auto_approval_delay_in_seconds = fields.Int(missing=None)
    assignment_duration_in_seconds = fields.Int(required=True,
                                                validate=validate.Range(min=30, max=31536000))
    lifetime_in_seconds = fields.Int(required=True,
                                     validate=validate.Range(min=30, max=31536000))
    reward = fields.Int(required=True, validate=validate.Range(min=1, max=1000))
    qualification_requirements = fields.List(fields.Dict(), missing=[])
    args = fields.Dict(keys=fields.Str())
    qualification_for = fields.UUID(required=False)


class BatchHitTaskSchema(Schema):
    app_id = fields.UUID(required=True)
    hit_args = fields.Nested('CreateHitSchema')


class BatchHitSchema(Schema):
    hits = fields.List(fields.Nested('BatchHitTaskSchema'))


class Assignment(db.Model):
    __table_args__ = (UniqueConstraint('task_id', 'assignment_id', 'worker_id'), )

    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    task_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey(Task.id), nullable=False)
    assignment_id = db.Column(db.String, nullable=True)
    worker_id = db.Column(db.String, nullable=True)
    app_version = db.Column(db.String, nullable=True, default=None)
    status = db.Column(pg.ENUM('Abandoned', 'Returned', 'Accepted', 'Submitted', 'Approved',
                               'Rejected', name='assignment_status_enum'), default='Accepted',
                       nullable=False)
    answer = db.Column(pg.JSONB, nullable=True)

    task = db.relationship('Task', backref=db.backref('assignments', lazy=True))

    sortable_fields = {
        'created': created
    }
    default_sort = 'created'
    default_desc = True

    @property
    def status_changes(self):
        return AssignmentStatusChange.query.filter_by(assignment_id=self.id) \
                                           .order_by('occurred').all()


# TODO: How do leases get garbage collected?
class KwitLease(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    task_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey(Task.id), nullable=False, unique=True)
    assignment_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey(Assignment.id), nullable=False)
    user_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey(User.id))

    task = db.relationship('Task', backref=db.backref('leases', lazy=True))
    assignment = db.relationship('Assignment', backref=db.backref('leases', lazy=True))
    user = db.relationship('User', backref=db.backref('leases', lazy=True))


class KwitAssignmentLeaseSchema(BaseSchema):
    assignment = fields.Nested('AssignmentSchema')
    user = fields.Nested('UserSchema')


event.listen(Assignment.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE VIEW view_assignment_status_times AS (
                SELECT
                assignment_id, accepted, submitted, returned, abandoned, approved,
                EXTRACT(epoch from (submitted - accepted)) as tt_submitted,
                EXTRACT(epoch from (approved - accepted))  as tt_approved,
                EXTRACT(epoch from (returned - accepted))  as tt_returned,
                EXTRACT(epoch from (abandoned - accepted))  as tt_abandoned
                FROM
                CROSSTAB('
                    SELECT cross_product.assignment_id, cross_product.status, s.occurred
                    FROM (
                    SELECT DISTINCT assignment_id, statuses.status
                    FROM assignment_status_change
                    CROSS JOIN (
                        SELECT DISTINCT status FROM assignment_status_change
                        ) AS statuses
                    ) AS cross_product
                    LEFT OUTER JOIN assignment_status_change AS s
                    ON cross_product.assignment_id = s.assignment_id AND
                       cross_product.status = s.status
                    ORDER BY 1, 2
                ') AS (assignment_id UUID,
                    abandoned     TIMESTAMP,
                    returned      TIMESTAMP,
                    accepted      TIMESTAMP,
                    submitted     TIMESTAMP,
                approved TIMESTAMP));
"""))

event.listen(Assignment.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE VIEW view_assignment_detail AS (
                SELECT vast.assignment_id,
                    vast.accepted,
                    vast.submitted,
                    vast.returned,
                    vast.abandoned,
                    vast.approved,
                    vast.tt_submitted,
                    vast.tt_approved,
                    vast.tt_returned,
                    vast.tt_abandoned,
                    a.created AS assignment_created,
                    a.status AS assignment_status,
                    a.worker_id,
                    a.app_version,
                    aa.id AS app_id,
                    aa.name AS app_name,
                    t.id AS task_id,
                    t.title AS task_title,
                    t.max_assignments,
                    t.status AS task_status,
                    t.reward,
                    t.sandbox,
                    t.assignment_duration_in_seconds
                FROM view_assignment_status_times vast,
                    assignment a,
                    annotation_app aa,
                    task t
                WHERE a.task_id = t.id AND
                        aa.id = t.annotation_app_id AND
                        vast.assignment_id = a.id);
"""))

event.listen(Assignment.__table__, 'after_create',
             DDL("""
            CREATE INDEX ix_assignment_created ON assignment (created DESC)
            """))
event.listen(Assignment.__table__, 'after_create',
             DDL("""
            CREATE INDEX ix_assignment_status ON assignment (status)
            """))


class AssignmentStatusChange(db.Model):
    __table_args__ = (UniqueConstraint('assignment_id', 'status'),)

    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    occurred = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    assignment_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey(Assignment.id), nullable=False)
    status = db.Column(pg.ENUM('Abandoned', 'Returned', 'Accepted', 'Submitted', 'Approved',
                               'Rejected', name='assignment_status_enum'), default='Accepted',
                       nullable=False)

    assignment = db.relationship('Assignment')


# IMPORTANT NOTE:
#
# This logic is replicated in
# stumpf/mturk_utils.py:handle_mturk_assignment_event(). If you update
# this logic, you MUST check that function and update the logic there!

# Why? Because of the race condition between creating Assignments
# through the authenticate endpoint, and the SNS webook, we must use
# insert().on_conflict_do_update() logic.  This prevents us from using
# the ORM AssignmentStatusChange model for updating the assignment
# status in that function so this code is NOT run in that code-path.
@event.listens_for(AssignmentStatusChange, 'after_insert')
def assignment_status_change_hook(mapper, connection, assignment_status_change):
    from flask import current_app as flask_app
    if assignment_status_change.status in ['Submitted']:
        celery_task_ids = []
        for ep in cached_iter_entry_points('stumpf.pipeline.tasks.assignment_status_change'):
            try:
                flask_app.logger.info(
                    f'Calling assignment_status_change hook: {ep.name} on '
                    '{assignment_status_change.assignment_id}')
                func = ep.load()
                # TODO: Better error checking here
                async_res = func.delay(assignment_status_change.id,
                                       assignment_status_change.assignment_id,
                                       assignment_status_change.status)
                celery_task_ids.append(str(async_res.id))

            except ImportError:
                flask_app.logger.error(f'Failed to load assignment_status_change hook: {ep.name}')

        if len(celery_task_ids):
            # Cannot use db.session or ORM modesl in after_insert hooks. See:
            # https://docs.sqlalchemy.org/en/13/orm/events.html#sqlalchemy.orm.events.MapperEvents.after_insert
            connection.execute(insert(Log.__table__).values(
                category='assignment-status-change-event',
                message={
                    'status_change': str(assignment_status_change.id),
                    'celery_ids': celery_task_ids}))
        else:
            flask_app.logger.error('Expected at least one assignment_status_change handler')


event.listen(AssignmentStatusChange.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE FUNCTION update_assignment_status()
             RETURNS TRIGGER AS
             $BODY$
             BEGIN
                UPDATE assignment SET status = (
                    SELECT status FROM assignment_status_change
                    WHERE assignment_id = NEW.assignment_id
                    ORDER BY occurred DESC LIMIT 1)
                WHERE id = NEW.assignment_id;
             RETURN NEW;
             END;
             $BODY$
             LANGUAGE plpgsql VOLATILE;"""))
event.listen(AssignmentStatusChange.__table__, 'after_create',
             DDL("""
             CREATE TRIGGER update_assignment_status
             AFTER INSERT OR UPDATE ON assignment_status_change
             FOR EACH ROW EXECUTE PROCEDURE update_assignment_status();"""))


event.listen(AssignmentStatusChange.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE VIEW view_assignment_status_change_detail AS (
                SELECT a.id,
                    a.created AS assignment_created,
                    asc_.status AS assignment_status,
                    a.worker_id,
                    a.app_version,
                    aa.id AS app_id,
                    aa.name AS app_name,
                    t.id AS task_id,
                    t.title AS task_title,
                    t.max_assignments,
                    t.status AS task_status,
                    t.reward,
                    t.sandbox,
                    t.assignment_duration_in_seconds
                FROM assignment_status_change asc_,
                    assignment a,
                    annotation_app aa,
                    task t
                WHERE a.task_id = t.id AND
                        aa.id = t.annotation_app_id AND
                        asc_.assignment_id = a.id);
"""))


class AssignmentStatusChangeSchema(BaseSchema):
    __model__ = AssignmentStatusChange

    created = fields.DateTime(dump_only=True)
    occurred = fields.DateTime()
    assignment_id = fields.Str()
    status = fields.Str()


class AssignmentSchema(BaseSchema):
    __model__ = Assignment

    task = fields.Nested('TaskSchema', dump_only=True)
    assignment_id = fields.Str()
    worker_id = fields.Str()
    status = fields.Str()
    app_version = fields.Str(required=True)
    status_changes = fields.List(fields.Nested('AssignmentStatusChangeSchema'), dump_only=True)
    answer = fields.Dict(required=False)


class QualificationType(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    user_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('user.id'), nullable=False)
    user = db.relationship('User', backref=db.backref('qualification_types', lazy=True))
    mturk_qt_id = db.Column(db.String, nullable=True, unique=True)
    sandbox = db.Column(pg.BOOLEAN, nullable=False, default=False)
    name = db.Column(db.String, nullable=False, unique=True)
    description = db.Column(db.String, nullable=True)
    keywords = db.Column(db.String, nullable=True)
    status = db.Column(pg.ENUM('Active', 'Inactive', name='qualification_type_status_enum'),
                       nullable=False, default='Active')


class QualificationTypeSchema(BaseSchema):
    __model__ = QualificationType

    mturk_qt_id = fields.Str()
    sandbox = fields.Boolean()
    name = fields.Str()
    description = fields.Str()
    keywords = fields.Str()
    status = fields.Str()


class CreateQualificationTypeSchema(Schema):
    sandbox = fields.Boolean(missing=True)
    name = fields.Str(required=True)
    description = fields.Str(required=True)
    keywords = fields.Str()
    status = fields.Str(missing='Active', validate=validate.OneOf(['Active', 'Inactive']))


class Log(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    category = db.Column(db.String, nullable=False, index=True)
    message = db.Column(pg.JSONB, nullable=False, default=lambda: {})


# This view assumes only 1 event per SNS message, which isn't a guarantee - though
# it holds true for all messages received to this point.
event.listen(Log.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE VIEW view_sns_events AS (
             SELECT id AS log_id,
                    to_timestamp(message->'Message'->'Events'->0->>'EventTimestamp',
                                 'YYYY-MM-DD"T"HH24:MI:SS"Z"')::timestamp at time zone 'UTC'
                                 AS occurred,
                    message->'Message'->'Events'->0->>'HITId' as hit_id,
                    message->'Message'->'Events'->0->>'AssignmentId' as assignment_id,
                    message->'Message'->'Events'->0->>'EventType' as event_type,
                    message->'Message'->'Events'->0->>'WorkerId' as worker_id
             FROM log
             WHERE category = 'raw-sns-message'
             ORDER BY occurred, created)
             """))


class LogSchema(BaseSchema):
    __model__ = Log

    category = fields.Str(required=True)
    message = fields.Dict(required=True)


class Feedback(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    ip = db.Column(db.String, nullable=False)
    user_agent = db.Column(db.String, nullable=False)
    args = db.Column(pg.JSONB, nullable=False, default=lambda: {})


class FeedbackSchema(BaseSchema):
    __model__ = Feedback

    ip = fields.Str(required=True)
    user_agent = fields.Str(required=True)
    args = fields.Dict(required=True)


class Worker(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    mturk_id = db.Column(db.String, nullable=False, unique=True)
    meta = db.Column(pg.JSONB, nullable=False)


class WorkerQualification(db.Model):
    __table_args__ = (UniqueConstraint('worker_id', 'qualification_type_id'), )

    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    worker_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('worker.id'), nullable=False)
    qualification_type_id = db.Column(
        pg.UUID(as_uuid=True), db.ForeignKey('qualification_type.id'), nullable=False)
    qualification_value = db.Column(db.Integer, nullable=False)

    worker = db.relationship('Worker', backref=db.backref('qualifications'), lazy=True)
    qualification_type = db.relationship('QualificationType')


class WorkerQualificationSchema(BaseSchema):
    __model__ = WorkerQualification
    worker_id = fields.UUID(required=True, load_only=True)
    qualification_type_id = fields.UUID(required=True, load_only=True)
    qualification_value = fields.Int(required=True)

    qualification = fields.Nested('QualificationTypeSchema')


class CreateWorkerQualificationSchema(Schema):
    score = fields.Int(missing=1)
    notify = fields.Boolean(missing=True)


class WorkerSchema(BaseSchema):
    __model__ = Worker
    mturk_id = fields.Str(required=True)
    meta = fields.Dict(missing=dict)

    qualifications = fields.List(fields.Nested('WorkerQualificationSchema', dump_only=True))
