from base64 import b64decode
from datetime import datetime
from typing import Dict, Optional

import boto3
from botocore.exceptions import ClientError
from flask import current_app
from jinja2 import Template
import requests
from sqlalchemy.dialects.postgresql import insert

from stumpf import cached_iter_entry_points
from stumpf.models import Assignment, AssignmentStatusChange, db, Log, QualificationType, \
    Task, TaskStatusChange, Worker, WorkerQualification

BAN_LIST_QUALIFICATION_ID = '3X4G950TG9YGXAU5AKOMBA7ROUJP1G'
UNBAN_MESSAGE = """
You have been removed from our internal ban list.
You are now eligible to perform our tasks.
"""
external_question_tpl = Template("""<?xml version="1.0"?><ExternalQuestion xmlns="http://mechanicalturk.amazonaws.com/AWSMechanicalTurkDataSchemas/2006-07-14/ExternalQuestion.xsd"><ExternalURL>{{ app_url }}?{{ query_string|escape }}</ExternalURL><FrameHeight>800</FrameHeight></ExternalQuestion>""")  # noqa


def question_xml(**kwargs):
    return external_question_tpl.render(**kwargs)


def generate_message(data: Dict) -> str:
    """
    Takes an SNS data value and generates the signable message.

    This is according to the specification detailed here:
    https://docs.aws.amazon.com/sns/latest/dg/SendMessageToHttp.verify.signature.html
    """
    message_keys = {'Notification': ['Message',
                                     'MessageId',
                                     'Subject',
                                     'Timestamp',
                                     'TopicArn',
                                     'Type'],
                    'SubscriptionConfirmation': ['Message',
                                                 'MessageId',
                                                 'SubscribeURL',
                                                 'Timestamp',
                                                 'Token',
                                                 'TopicArn',
                                                 'Type'],
                    'UnsubscribeConfirmation': ['Message',
                                                'MessageId',
                                                'SubscribeURL',
                                                'Timestamp',
                                                'Token',
                                                'TopicArn',
                                                'Type']}

    keys = message_keys[data['Type']]
    message = ''

    for key in keys:
        message += f'{key}\n{data[key]}\n'

    return message


def verify_signature(data: Dict) -> bool:
    from M2Crypto import X509
    try:
        message = generate_message(data)
        signature = b64decode(data['Signature'])

        cert = X509.load_cert_string(requests.get(data['SigningCertURL']).text)
        pubkey = cert.get_pubkey()
        pubkey.reset_context(md='sha1')
        pubkey.verify_init()
        pubkey.verify_update(message.encode())

        return True if pubkey.verify_final(signature) == 1 else False
    except Exception:
        current_app.logger.warning('Failed when attempting to verify signature')
        return False


def handle_mturk_assignment_event(event: Dict):
    task = Task.query.filter_by(hit_id=event['HITId']).first()

    if not task:
        current_app.logger.info(f'Received event with unknown task, event: {event}')
        current_app.logger.warning(f'Received event with unknown task')
        raise Exception('Received event with unknown task')

    event['EventTimestamp'] = datetime.strptime(event['EventTimestamp'], '%Y-%m-%dT%H:%M:%SZ')

    values = {
        'task_id': task.id,
        'assignment_id': event['AssignmentId'],
        'worker_id': event['WorkerId'],
        'status': 'Accepted'
    }

    try:
        """
        Since this endpoint can be processed in parallel, and other endpoints will have the ability
        to create assignments and update their status - this is wrapped in a transaction. Inserting
        a status change will activate a trigger to store the status in the assignment table as well.
        """
        db.session.execute(
            insert(Assignment.__table__).values(**values)
            .on_conflict_do_nothing(constraint='uq_assignment_task_id'))

        assignment = Assignment.query.filter_by(task=task,
                                                assignment_id=event['AssignmentId'],
                                                worker_id=event['WorkerId']).first()

        assignment_id, assignment_status = assignment.id, assignment.status
        sns_event_status = event['EventType'].replace('Assignment', '')

        status_change_result = db.session.execute(
            insert(AssignmentStatusChange.__table__).values(
                assignment_id=assignment.id,
                occurred=event['EventTimestamp'],
                status=sns_event_status)
            .on_conflict_do_update(constraint='uq_assignment_status_change_assignment_id',
                                   set_={'occurred': event['EventTimestamp']}))
        db.session.commit()

        # Only trigger if the event is a submitted event, and the
        # assignment status change is different than the assignment's
        # previous status (e.g.  we actually inserted, instead of
        # on_conflict updating).
        #
        # NOTE: For now this logic is replicated in
        # stumpf/models.py::assignment_status_change_hook - See note
        # there. Changes here need to be made there as well.
        if sns_event_status == 'Submitted' and assignment_status != sns_event_status:
            assignment_status_change_id = str(status_change_result.inserted_primary_key[0])
            celery_task_ids = []

            for ep in cached_iter_entry_points('stumpf.pipeline.tasks.assignment_status_change'):
                try:
                    current_app.logger.info(
                        f'Calling assignment_status_change hook: {ep.name} on {assignment_id}')
                    func = ep.load()
                    # TODO: Better error checking here
                    async_res = func.delay(assignment_status_change_id,
                                           assignment_id, sns_event_status)
                    celery_task_ids.append(str(async_res.id))
                except ImportError:
                    current_app.logger.error(
                        f'Failed to load assignment_status_change hook: {ep.name}')

            if len(celery_task_ids):
                db.session.add(Log(category='assignment-status-change-event',
                                   message={
                                       'status_change': assignment_status_change_id,
                                       'celery_ids': celery_task_ids}))
                db.session.commit()
            else:
                current_app.logger.error('Expected at least one assignment_status_change handler')

    except Exception:
        db.session.rollback()
        current_app.logger.info(f'Failed to create/update assignment, event: {event}')
        current_app.logger.warning(f'Failed to create/update assignment')
        raise


def handle_mturk_hit_event(event: Dict):
    task = Task.query.filter_by(hit_id=event['HITId']).first()

    if not task:
        if event['EventType'] == 'HITDisposed':
            # It's possible we deleted the task database side, and received a disposal event long
            # after deletion. If we deleted the task in our database, there's no need to track
            # or warn about this.
            current_app.logger.info('Received HITDisposed event for unknown task.', event)
        else:
            # If a different kind of event comes in for a task that doesn't exist, something
            # went wrong
            current_app.logger.info('Received event for unknown task', event)
            current_app.logger.warning('Received event for unknown task')
        return
    elif event['EventType'] == 'HITCreated':
        return  # sanity check to ensure the HIT exists

    status_change = TaskStatusChange(task=task)
    if event['EventType'] == 'HITReviewable':
        status_change.status = 'Reviewable'
    elif event['EventType'] == 'HITExpired':
        # From MT blog:
        # "HITs become Unassignable when either the last assignment for that HIT is accepted by
        #  a Worker or when the HIT expires."
        #
        # We are ignoring that because then there is no way to distinguish expired hits.  Instead,
        # we set a custom Expired status.
        status_change.status = 'Expired'
    elif event['EventType'] == 'HITDisposed':
        status_change.status = 'Disposed'
    else:
        current_app.logger.info('Received unknown event type', event)
        current_app.logger.warning('Received unknown event type')
        return

    try:
        status_change.occurred = event['EventTimestamp']
        db.session.add(status_change)
        db.session.commit()

        if status_change.status in ['Reviewable', 'Expired', 'Disposed']:
            current_app.statsd.gauge('hits.active', -1, delta=True)
    except Exception:
        current_app.logger.info('Failed to update task', event)
        current_app.logger.warning('Failed to update task')
        raise


def create_hit(client, assignment_duration: int, reward: int, title: str,
               description: str, keywords: str, max_assignments: int,
               lifetime_in_seconds: int, question: str,
               auto_approval_delay_in_seconds: Optional[int],
               qualification_requirements: Optional[Dict]) -> Dict:
    hit_type = create_hit_type(client, assignment_duration,
                               reward, title, description, keywords, auto_approval_delay_in_seconds,
                               qualification_requirements)

    return client.create_hit_with_hit_type(HITTypeId=hit_type['HITTypeId'],
                                           MaxAssignments=max_assignments,
                                           LifetimeInSeconds=lifetime_in_seconds,
                                           Question=question)


def create_hit_type(client, assignment_duration: int, reward: int, title: str,
                    description: str, keywords: str,
                    auto_approval_delay_in_seconds: Optional[int],
                    qualification_requirements: Optional[Dict]) -> Dict:
    kwargs = {}

    if auto_approval_delay_in_seconds:
        kwargs['AutoApprovalDelayInSeconds'] = auto_approval_delay_in_seconds

    if qualification_requirements:
        kwargs['QualificationRequirements'] = qualification_requirements

    hit_type = client.create_hit_type(AssignmentDurationInSeconds=assignment_duration,
                                      Reward=str(reward / 100),
                                      Title=title,
                                      Description=description,
                                      Keywords=keywords,
                                      **kwargs)

    client.update_notification_settings(
        HITTypeId=hit_type['HITTypeId'],
        Notification={
            'Destination': current_app.config['MTURK_SNS_ARN'],
            'Transport': 'SNS',
            'Version': '2014-08-15',
            'EventTypes': [
                'AssignmentAccepted',
                'AssignmentAbandoned',
                'AssignmentReturned',
                'AssignmentSubmitted',
                'AssignmentRejected',
                'AssignmentApproved',
                'HITCreated',
                'HITExpired',
                'HITReviewable',
                'HITDisposed']},
        Active=True)

    return hit_type


def get_worker_qualification_score(client, worker_id: str, qualification_id: str) -> Optional[int]:
    """Return a worker's score for a given qualification type id.

    This method will return ``None`` when the user does not have the qualification,
    otherwise it will return the integer value.

    **NOTE** this will only work for qualification types created by us.  Default
    types (like ``00000000000000000040``) cannot be queried and will return None.
    """
    try:
        qualification = client.get_qualification_score(
            QualificationTypeId=qualification_id,
            WorkerId=worker_id
        )['Qualification']
    except ClientError:
        return

    assert qualification['QualificationTypeId'] == qualification_id
    if qualification['Status'] == 'Granted':
        return qualification['IntegerValue']


def set_worker_qualification_score(client, worker_id: str, qualification_id: str, score: int,
                                   notify=True):
    qualification = QualificationType.query.filter_by(
        mturk_qt_id=qualification_id).first_or_404()
    worker = Worker.query.filter_by(
        mturk_id=worker_id).first_or_404()

    db.session.execute(
        insert(WorkerQualification.__table__)
        .values(
            worker_id=worker.id,
            qualification_type_id=qualification.id,
            qualification_value=score
        ).on_conflict_do_update(
            constraint='uq_worker_qualification_worker_id',
            set_={'qualification_value': score})
    )

    try:
        resp = client.associate_qualification_with_worker(
            QualificationTypeId=qualification_id,
            WorkerId=worker_id,
            IntegerValue=score,
            SendNotification=notify
        )
        db.session.commit()
    except Exception:
        db.session.rollback()
        raise

    return resp


def remove_worker_qualification_score(client, worker_id: str, qualification_id: str, reason: str):
    # avoid circular imports
    from stumpf.models import db, QualificationType, Worker, WorkerQualification

    qualification = QualificationType.query.filter_by(
        mturk_qt_id=qualification_id).first_or_404()
    worker = Worker.query.filter_by(
        mturk_id=worker_id).first_or_404()
    WorkerQualification.query.filter_by(
        worker_id=worker.id,
        qualification_type_id=qualification.id
    ).delete()
    resp = None

    try:
        resp = client.disassociate_qualification_from_worker(
            WorkerId=worker_id,
            QualificationTypeId=qualification_id,
            Reason=reason
        )
        db.session.commit()
    except Exception:
        db.session.rollback()

    return resp


def set_qualification_score_from_assignment(assignment, score):
    """Associate a worker with a score from a qualification assignment."""
    if assignment.task.qualification_for_id is None:
        # this is not a qualification task
        return

    qualification_id = assignment.task.qualification_for.mturk_qt_id
    worker_id = assignment.worker_id
    session = boto3.session.Session()

    if assignment.task.sandbox:
        mturk = session.client(
            'mturk', endpoint_url=current_app.config['MTURK_SANDBOX_ENDPOINT_URL'])
    else:
        mturk = session.client(
            'mturk', endpoint_url=current_app.config['MTURK_PRODUCTION_ENDPOINT_URL'])
    set_worker_qualification_score(mturk, worker_id, qualification_id, score)


# These methods will not work on clients connection to sandbox endpoints because the
# qualification types are different.
def is_worker_banned(client, worker_id: str) -> bool:
    return get_worker_qualification_score(client, worker_id, BAN_LIST_QUALIFICATION_ID) is not None


def ban_worker(client, worker_id: str):
    return set_worker_qualification_score(client, worker_id, BAN_LIST_QUALIFICATION_ID, 1, False)


def unban_worker(client, worker_id: str):
    return remove_worker_qualification_score(client, worker_id, BAN_LIST_QUALIFICATION_ID,
                                             UNBAN_MESSAGE)
