from functools import wraps

from flask import current_app

from stumpf.models import Assignment, db, Task


class PipelineStepFilter:
    def __init__(self, app_id, types):
        self.app_id = str(app_id)
        self.types = types

    def _task_filter(self, task):
        return (
            not task.sandbox
            and str(task.annotation_app_id) == self.app_id
            and 'no-pipeline' not in (task.keywords or '').split(',')
            and task.type_ in self.types
        )

    def should_run_on_assignment(self, assignment):
        if assignment is None:
            raise Exception('Could not load assignment')
        if assignment.status != 'Submitted':
            # This would indicate some sort of race condition in the
            # pipeline processing.  Just log it for future reference.
            current_app.logger.info(
                f'Pipeline run on assignment {assignment.id} with status {assignment.status}'
            )
        return self._task_filter(assignment.task)

    def should_run_on_task(self, task):
        if task is None:
            raise Exception('Could not load task')
        if task.status != 'Reviewable':
            # This would indicate some sort of race condition in the
            # pipeline processing.  Just log it for future reference.
            current_app.logger.info(
                f'Pipeline run on task {task.id} with status {task.status}'
            )
        return self._task_filter(task)


def assignment_filter(app_id, types=('HIT', 'KWIT')):
    filter = PipelineStepFilter(app_id, types)

    def decorator(func):
        @wraps(func)
        def wrapper(status_change_id, assignment_id, status):
            assignment = Assignment.query.get(assignment_id)
            if filter.should_run_on_assignment(assignment):
                try:
                    current_app.logger.info(
                        f'Running adjudication for {app_id} on assignment {assignment_id}')
                    return func(assignment)
                except Exception:
                    db.session.rollback()
                    raise
        return wrapper
    return decorator


def task_filter(app_id, types=('HIT', 'KWIT')):
    filter = PipelineStepFilter(app_id, types)

    def decorator(func):
        @wraps(func)
        def wrapper(status_change_id, task_id, status):
            if status == 'Expired':
                return
            task = Task.query.get(task_id)
            if filter.should_run_on_task(task):
                try:
                    current_app.logger.info(
                        f'Running adjudication for {app_id} on task {task_id}')
                    return func(task)
                except Exception:
                    db.session.rollback()
                    raise
        return wrapper
    return decorator
