from flask import Blueprint, jsonify, request

from stumpf.middleware import google_auth_middleware
from stumpf.models import AnnotationApp, AnnotationAppSchema, db
from stumpf.pagination import paginated_response
from stumpf.sorting import sorted_query


annotation_app_bp = Blueprint('annotation_app', __name__)
annotation_app_bp.before_request(google_auth_middleware)

annotation_app_schema = AnnotationAppSchema()


@annotation_app_bp.route('/annotation-app', methods=['POST'])
def create_annotation_app():
    """
    ---
    post:
        summary: Create a new annotation app
        description: Create a new annotation app
        tags:
            - AnnotationApp
        parameters:
            - in: json
              schema: AnnotationAppSchema
        responses:
            201:
                schema: AnnotationAppSchema
    """
    try:
        annotation_app = annotation_app_schema.load(request.json)

        if AnnotationApp.query.filter_by(name=annotation_app.name,
                                         repo_url=annotation_app.repo_url,
                                         url=annotation_app.url).first():
            return jsonify({
                'error': 'Annotation app already exists'
            }), 409

        db.session.add(annotation_app)
        db.session.commit()
        return jsonify(annotation_app_schema.dump(annotation_app)), 201
    except Exception:
        db.session.rollback()
        raise


@annotation_app_bp.route('/annotation-app', methods=['GET'])
def list_annotation_app():
    """
    ---
    get:
        summary: List annotation apps
        description: List annotation apps
        tags:
            - AnnotationApp
        parameters:
          - name: sort
            in: query
            type: string
          - name: desc
            in: query
            type: boolean
        responses:
            200:
                schema:
                    type: array
                    items: AnnotationAppSchema
    """
    query = sorted_query(AnnotationApp)
    return paginated_response(query, lambda items: annotation_app_schema.dump(items, many=True))


@annotation_app_bp.route('/annotation-app/<uuid:annotation_app_id>', methods=['GET'])
def get_annotation_app(annotation_app_id):
    """
    ---
    get:
        summary: Retrieve a single annotation app
        description: Retrieve a single annotation app
        tags:
            - AnnotationApp
        parameters:
          - name: annotation_app_id
            description: UUID of the annotation app
            in: path
            type: string
        responses:
            200:
                schema: AnnotationAppSchema
            404:
                description: Annotation app does not exist
    """
    annotation_app = AnnotationApp.query.get_or_404(annotation_app_id)
    return jsonify(annotation_app_schema.dump(annotation_app))


@annotation_app_bp.route('/annotation-app/<uuid:annotation_app_id>', methods=['DELETE'])
def delete_annotation_app(annotation_app_id):
    """
    ---
    delete:
        summary: Delete an annotation app
        description: Delete an annotation app
        tags:
            - AnnotationApp
        parameters:
          - name: annotation_app_id
            description: UUID of the annotation app
            in: path
            type: string
        responses:
            204:
                description: Annotation app deleted
    """
    try:
        AnnotationApp.query.filter_by(id=annotation_app_id).delete()
        db.session.commit()
        return '', 204

    except Exception:
        db.session.rollback()
        raise
