from flask import Blueprint, jsonify, request
from flask_cors import CORS
from flask_login import current_user
from marshmallow import fields

from stumpf.kwit_utils import assignment_url_from_lease
from stumpf.middleware import google_auth_middleware
from stumpf.models import Assignment, AssignmentSchema, AssignmentStatusChange, \
    db, KwitLease, Task, TaskStatusChange
from stumpf.pagination import paginated_response
from stumpf.sorting import sorted_query


assignment_bp = Blueprint('assignment', __name__)
assignment_bp.before_request(google_auth_middleware)
CORS(assignment_bp, expose_headers=['resource-count'])

assignment_schema = AssignmentSchema()


@assignment_bp.route('/assignment', methods=['GET'])
def list_assignments():
    query = Assignment.query.filter(
        Assignment.status.in_(['Approved', 'Submitted'])
    )
    if 'task_id' in request.args:
        query = query.filter_by(task_id=request.args['task_id'])
    if 'worker_id' in request.args:
        query = query.filter_by(worker_id=request.args['worker_id'])

    query = sorted_query(Assignment, query)
    return paginated_response(query, lambda items: assignment_schema.dump(items, many=True))


@assignment_bp.route('/assignment/open/mine', methods=['GET'])
def list_my_open_assignments():
    class AssignmentWithUrl(AssignmentSchema):
        assignment_url = fields.Url(dump_only=True)
    schema = AssignmentWithUrl()

    assignments = db.session.query(Assignment, KwitLease) \
                    .filter(Assignment.id == KwitLease.assignment_id) \
                    .filter(KwitLease.user_id == current_user.id) \
                    .filter(Assignment.status == 'Accepted') \
                    .order_by(Assignment.created) \
                    .all()

    for (assignment, lease) in assignments:
        assignment.assignment_url = assignment_url_from_lease(lease)

    return jsonify(schema.dump((x[0] for x in assignments), many=True))


@assignment_bp.route('/assignment/<uuid:assignment_id>/abandon', methods=['POST'])
def abandon_assignemnt(assignment_id):
    assignment = Assignment.query.get_or_404(assignment_id)
    lease = KwitLease.query \
        .filter_by(assignment_id=assignment_id) \
        .filter_by(user_id=current_user.id).first()
    if lease is None:
        return jsonify({'error': 'No assignment with ID is currently assigned to you'}), 400
    try:
        status_change = AssignmentStatusChange(assignment_id=assignment_id,
                                               status='Abandoned')
        task_status_change = TaskStatusChange(task=assignment.task, status='Assignable')
        db.session.add_all([status_change, task_status_change, lease])
        db.session.delete(lease)
        db.session.commit()
        return jsonify(''), 204
    except Exception:
        db.session.rollback()
        raise


@assignment_bp.route('/assignment/<uuid:assignment_id>', methods=['GET'])
def get_assignment(assignment_id):
    assignment = Assignment.query.get_or_404(assignment_id)
    return jsonify(assignment_schema.dump(assignment))


@assignment_bp.route('/assignment', methods=['POST'])
def create_assignment():
    try:
        assignment = assignment_schema.load(request.json)

        if not Task.query.get(assignment.task_id):
            return jsonify({
                'error': 'No such task.'
            }), 400

        db.session.add(assignment)
        db.session.commit()
        return jsonify(assignment_schema.dump(assignment)), 201
    except Exception:
        db.session.rollback()
        raise

    return jsonify(assignment_schema.dump(assignment))


# todo put vs patch
@assignment_bp.route('/assignment/<uuid:assignment_id>', methods=['PUT'])
def update_assignment(assignment_id):
    assignment = Assignment.query.get_or_404(assignment_id)

    if assignment.completed is not None:
        return jsonify({
            'error': 'Assignment is already completed'
        }), 409

    try:
        assignment_update_attrs = assignment_schema.load(request.json, only=('completed',))

        for attr, value in assignment_update_attrs.items():
            setattr(assignment, attr, value)

        db.session.add(assignment)
        db.session.commit()
        return jsonify(assignment_schema.dump(assignment)), 201
    except Exception:
        db.session.rollback()
        raise

    return jsonify(assignment_schema.dump(assignment))
