import celery
from flask import Blueprint, jsonify, request
from flask_login import current_user

from stumpf.celery.tasks import create_hit_task
from stumpf.decorators import handle_boto_client_errors
from stumpf.middleware import google_auth_middleware
from stumpf.models import (AnnotationApp, BatchHitSchema, db)


batch_bp = Blueprint('batch', __name__)
batch_bp.before_request(google_auth_middleware)

batch_hit_schema = BatchHitSchema()


def _validate_app_ids(app_ids):
    # Here we build a hash of app_id -> app_url based on the app_id's
    # that were passed in. This hash will ONLY contain ids that are
    # actually in the database.  If there are ids that were passed in
    # (i.e. from the app_ids variable) then these will NOT be in the
    # hash.  We use this fact to do a set difference and return a list
    # of app ids that were passed in, but are not in the database.
    # (i.e. errors).

    app_id_url_hash = {r[0]: r[1] for r in
                       db.session.query(AnnotationApp.id, AnnotationApp.url)
                       .filter(AnnotationApp.id.in_(app_ids)).all()}

    db_app_ids = {key for key in app_id_url_hash.keys()}

    # Note that in addition to invalid app id's that we received we
    # also return the app_id_url_hash which is needed later on in
    # processing to allow us to pass the app_url to create_hit_task
    # for HIT creation on mturk.
    return [str(_id) for _id in app_ids - db_app_ids], app_id_url_hash


@batch_bp.route('/task/batch/hit', methods=['POST'])
@handle_boto_client_errors
def create_batch_hits():
    batch_list = batch_hit_schema.load(request.json)

    # Validate the application ids that were passed in
    app_errs, app_id_url_hash = \
        _validate_app_ids({t['app_id'] for t in batch_list['hits']})

    if app_errs:
        return 'Error: The following annotation app ids were not found: {}\n'.format(
            ','.join(app_errs)), 422

    task_signatures = []

    # Create a list of task signatures
    for task in batch_list['hits']:
        app_url = app_id_url_hash[task['app_id']]

        task_signatures.append(
            create_hit_task.s(
                str(current_user.get_id()),
                str(task['app_id']),
                app_url,
                task['hit_args']))

    # Fire off the group
    group_result = celery.group(task_signatures).apply_async()
    group_result.save()

    return jsonify(group_result.id), 202
