from datetime import datetime, timedelta

from flask import Blueprint, current_app, g, jsonify, request
from flask_cors import CORS
from marshmallow import fields, post_load, Schema, validates, ValidationError
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.sql import select, update

from stumpf.decorators import public_endpoint
from stumpf.middleware import mturk_auth_middleware
from stumpf.models import (Assignment, AssignmentSchema, AssignmentStatusChange, db, FeedbackSchema,
                           KwitLease, LogSchema, MturkToken, MturkTokenSchema, Task,
                           TaskStatusChange, Worker)

feedback_schema = FeedbackSchema()
log_schema = LogSchema()
mturk_token_schema = MturkTokenSchema()

mturk_bp = Blueprint('mturk', __name__)
mturk_bp.before_request(mturk_auth_middleware)
CORS(mturk_bp)


# This is not thread-safe due to the use of self
class MturkAuthSchema(Schema):
    mturk_hit_id = fields.Str(required=True)
    mturk_assignment_id = fields.Str(required=True)
    app_version = fields.Str(required=True)
    mturk_worker_id = fields.Str(required=True)

    @validates('mturk_hit_id')
    def validate_hit_id(self, mturk_hit_id):
        self._task = Task.query.filter_by(hit_id=mturk_hit_id).first()

        if not self._task:
            raise ValidationError('Task does not exist.')

    @validates('mturk_assignment_id')
    def validate_assignment_id(self, mturk_assignment_id):
        if mturk_assignment_id == 'ASSIGNMENT_ID_NOT_AVAILABLE':
            raise ValidationError('Cannot login when previewing')

    @post_load
    def make_object(self, data, **kwargs):
        return data, self._task


@mturk_bp.route('/authenticate', methods=['POST'])
@public_endpoint
def authenticate():
    try:
        mturk_auth_schema = MturkAuthSchema()
        auth_data, task = mturk_auth_schema.load(request.json)

        sandbox = ' (sandbox) ' if task.sandbox else ' '

        current_app.logger.debug(f'MTurk{sandbox}authenticate called with auth data: {auth_data}')

        # Generate a row in the Worker database for this worker_id if none already exists
        db.session.execute(
            insert(Worker.__table__).values(
                mturk_id=auth_data['mturk_worker_id'],
                meta={})
            .on_conflict_do_nothing(
                constraint='uq_worker_mturk_id'))

        # Attempt to create the assignment, if it was created by SNS
        # already update the app_version (not provided by SNS).
        db.session.execute(
            insert(Assignment.__table__).values(
                task_id=task.id,
                assignment_id=auth_data['mturk_assignment_id'],
                app_version=auth_data['app_version'],
                worker_id=auth_data['mturk_worker_id'],
                status='Accepted')
            .on_conflict_do_update(constraint='uq_assignment_task_id',
                                   set_={'app_version': auth_data['app_version']}))

        # Fetch the assignment row so a status change can be inserted
        assignment_id = db.session.execute(select([Assignment.__table__])
                                           .with_only_columns([Assignment.__table__.c.id])
                                           .where(Assignment.task_id == task.id)
                                           .where(Assignment.assignment_id ==
                                                  auth_data['mturk_assignment_id'])
                                           .where(Assignment.worker_id ==
                                                  auth_data['mturk_worker_id'])
                                           .where(Assignment.app_version ==
                                                  auth_data['app_version'])
                                           .limit(1)).scalar()

        # Insert the status change, ignoring if it's already been logged for this assignment,
        # because that means the time came from SNS (which is the canonical event timestamp)
        db.session.execute(insert(AssignmentStatusChange.__table__).values(
            assignment_id=str(assignment_id), occurred=datetime.utcnow(), status='Accepted')
                           .on_conflict_do_nothing(
                               constraint='uq_assignment_status_change_assignment_id'))

        # Create token for the duration of the HIT assignment value
        expires_at = (datetime.utcnow() +
                      timedelta(seconds=task.assignment_duration_in_seconds))
        token = MturkToken(assignment_id=str(assignment_id), expires=expires_at)

        db.session.add(token)
        db.session.commit()

        return jsonify(mturk_token_schema.dump(token)), 201
    except Exception:
        db.session.rollback()
        raise


# This is not thread-safe due to the use of self
class KwitAuthSchema(Schema):
    assignment_id = fields.Str(required=True)
    app_version = fields.Str(required=True)

    @validates('assignment_id')
    def validate_assignment_id(self, assignment_id):
        self._assignment = Assignment.query.filter_by(id=assignment_id,
                                                      status='Accepted').first()

        if not self._assignment:
            raise ValidationError('Assignment does not exist or has been completed.')

    @post_load
    def make_object(self, data, **kwargs):
        return data, self._assignment


@mturk_bp.route('/authenticate/kwit', methods=['POST'])
@public_endpoint
def authenticate_kwit():
    try:
        kwit_auth_schema = KwitAuthSchema()
        auth_data, assignment = kwit_auth_schema.load(request.json)

        assignment.app_version = auth_data['app_version']
        expires_at = datetime.utcnow() + timedelta(days=7)
        token = MturkToken(assignment=assignment, expires=expires_at)

        db.session.add_all([assignment, token])
        db.session.commit()

        return jsonify(mturk_token_schema.dump(token)), 201
    except Exception:
        db.session.rollback()
        raise


@mturk_bp.route('/answer', methods=['POST'])
def submit_answer():
    try:
        schema = AssignmentSchema(only=('id', 'created', 'answer'))
        answer = schema.load(request.json).answer
        assignment = Assignment.query.get_or_404(g.assignment_id)

        if assignment.status != 'Accepted':
            return jsonify({
                'error': 'Assignment is no longer accepting answers.'
            }), 400

        db.session.execute(update(Assignment.__table__)
                           .where(Assignment.id == g.assignment_id)
                           .values(answer=answer))
        db.session.commit()

        return jsonify(schema.dump(assignment))
    except Exception:
        db.session.rollback()
        raise


@mturk_bp.route('/log', methods=['POST'])
def submit_log():
    try:
        log = log_schema.load(request.json)
        db.session.add(log)
        db.session.commit()
        return jsonify(log_schema.dump(log)), 201
    except Exception:
        db.session.rollback()
        raise


@mturk_bp.route('/feedback', methods=['POST'])
def submit_feedback():
    try:
        feedback = feedback_schema.load(
            dict({'ip': request.remote_addr,
                  'user_agent': request.headers.get('User-Agent', 'unknown')},
                 **request.json)
        )
        db.session.add(feedback)
        db.session.commit()
        return jsonify(feedback_schema.dump(feedback)), 201
    except Exception:
        db.session.rollback()
        raise


@mturk_bp.route('/submit', methods=['POST'])
def mark_submitted():
    try:
        assignment = Assignment.query.get_or_404(g.assignment_id)

        if assignment.status != 'Accepted':
            return jsonify({
                'error': 'Assignment can not be marked as submitted.'
            }), 400
        elif assignment.task.type_ != 'KWIT':
            return jsonify({
                'error': 'Only KWITs should be submitted this way.'
            }), 400

        status_change = AssignmentStatusChange(occurred=datetime.utcnow(), assignment=assignment,
                                               status='Submitted')
        task_status_change = TaskStatusChange(task=assignment.task, status='Reviewable')
        db.session.delete(KwitLease.query.filter_by(assignment=assignment).first())
        db.session.add_all([status_change, task_status_change])
        db.session.commit()
        current_app.statsd.gauge('kwits.active', -1, delta=True)
        return '', 200
    except Exception:
        db.session.rollback()
        raise
