from urllib.parse import urlencode

import boto3
from flask import Blueprint, current_app, jsonify, request
from flask_cors import CORS
from flask_login import current_user

from stumpf.celery import app
from stumpf.celery.tasks import create_hit_task
from stumpf.decorators import handle_boto_client_errors
from stumpf.filtering import filter_query_bool, filter_query_enum
from stumpf.kwit_utils import assignment_url_from_lease
from stumpf.middleware import google_auth_middleware
from stumpf.models import (AnnotationApp, Assignment, AssignmentSchema, AssignmentStatusChange,
                           CreateHitSchema, CreateKwitSchema, CreateTaskSchema, db,
                           KwitAssignmentLeaseSchema, KwitLease, Task, TaskSchema, TaskStatusChange)
from stumpf.mturk_utils import create_hit, question_xml
from stumpf.pagination import paginated_response
from stumpf.sorting import sorted_query


task_bp = Blueprint('task', __name__)
task_bp.before_request(google_auth_middleware)
CORS(task_bp, expose_headers=['resource-count'])
assignment_schema = AssignmentSchema()
task_schema = TaskSchema()
create_hit_schema = CreateHitSchema()
create_kwit_schema = CreateKwitSchema()
create_task_schema = CreateTaskSchema()
kwit_assignment_lease_schema = KwitAssignmentLeaseSchema()


@task_bp.route('/task/<uuid:app_id>/hit', methods=['POST'])
@handle_boto_client_errors
def create_task_hit_async(app_id):
    annotation_app = AnnotationApp.query.get_or_404(app_id)
    hit_args = create_hit_schema.load(request.json)

    current_app.logger.info(f'Trying to put task on queue: {app.app.pool.connection.as_uri()}')

    async_result = create_hit_task.delay(
        str(current_user.get_id()),
        str(annotation_app.id),
        str(annotation_app.url),
        hit_args)

    return jsonify(async_result.id), 202


@task_bp.route('/task/<uuid:app_id>/sync_hit', methods=['POST'])
@handle_boto_client_errors
def create_task_hit(app_id):
    annotation_app = AnnotationApp.query.get_or_404(app_id)
    hit_args = create_hit_schema.load(request.json)

    session = boto3.session.Session()
    if hit_args['sandbox']:
        mturk = session.client('mturk',
                               endpoint_url=current_app.config['MTURK_SANDBOX_ENDPOINT_URL'])
    else:
        mturk = session.client('mturk',
                               endpoint_url=current_app.config['MTURK_PRODUCTION_ENDPOINT_URL'])

    try:
        task = Task(annotation_app=annotation_app, user=current_user, type_='HIT',
                    sandbox=hit_args['sandbox'], title=hit_args['title'],
                    description=hit_args['description'], keywords=hit_args.get('keywords', ''),
                    max_assignments=hit_args['max_assignments'],
                    assignment_duration_in_seconds=hit_args['assignment_duration_in_seconds'],
                    reward=hit_args['reward'], args=hit_args.get('args', {}),
                    qualification_requirements=hit_args['qualification_requirements'],
                    qualification_for_id=hit_args.get('qualification_for'))
        status_change = TaskStatusChange(task=task, status='Assignable')
        db.session.add_all([task, status_change])
        db.session.commit()

        hit = create_hit(mturk, hit_args['assignment_duration_in_seconds'], hit_args['reward'],
                         hit_args['title'], hit_args['description'], hit_args.get('keywords', ''),
                         hit_args['max_assignments'], hit_args['lifetime_in_seconds'],
                         question_xml(app_url=annotation_app.url,
                                      query_string=urlencode({**hit_args.get('args', {})})),
                         hit_args['auto_approval_delay_in_seconds'],
                         hit_args['qualification_requirements'])

        current_app.logger.info(f'Created HIT {hit["HIT"]["HITId"]} for Task {task.id}')
        current_app.statsd.gauge('hits.active', 1, delta=True)

        task.hit_id = hit['HIT']['HITId']
        task.hit_group_id = hit['HIT']['HITGroupId']
        db.session.add(task)
        db.session.commit()

        return jsonify(task_schema.dump(task)), 201
    except Exception:
        db.session.rollback()
        raise


@task_bp.route('/task/<uuid:app_id>/kwit', methods=['POST'])
def create_task_kwit(app_id):
    annotation_app = AnnotationApp.query.get_or_404(app_id)
    kwit_args = create_kwit_schema.load(request.json)

    try:
        task = Task(annotation_app=annotation_app, user=current_user, type_='KWIT',
                    sandbox=kwit_args['sandbox'], title=kwit_args['title'],
                    description=kwit_args['description'], keywords=kwit_args.get('keywords', ''),
                    max_assignments=kwit_args['max_assignments'], args=kwit_args.get('args', {}))
        status_change = TaskStatusChange(task=task, status='Assignable')
        db.session.add_all([task, status_change])
        db.session.commit()
        current_app.statsd.gauge('kwits.active', 1, delta=True)

        return jsonify(task_schema.dump(task)), 201
    except Exception:
        db.session.rollback()
        raise


@task_bp.route('/task/kwit/<uuid:task_id>/lease', methods=['POST'])
def create_kwit_lease(task_id):
    task = Task.query.get_or_404(task_id)

    if task.type_ != 'KWIT':
        return jsonify({
            'error': 'Only KWIT tasks can be leased.'
        }), 400
    elif task.status != 'Assignable':
        return jsonify({
            'error': 'This KWIT cannot be leased.'
        }), 400

    try:
        assignment = Assignment(task=task,
                                worker_id=current_user.id,
                                status='Accepted')
        status_change = AssignmentStatusChange(assignment=assignment, status='Accepted')
        task_status_change = TaskStatusChange(task=assignment.task, status='Unassignable')
        lease = KwitLease(task=task, assignment=assignment, user=current_user)
        db.session.add_all([task, assignment, task_status_change, status_change, lease])
        db.session.commit()

        return jsonify({
            'url': assignment_url_from_lease(lease)
        })
    except Exception:
        db.session.rollback()
        raise


@task_bp.route('/task', methods=['GET'])
def list_tasks():
    query = sorted_query(Task)
    query = filter_query_enum(Task, 'type', query)
    query = filter_query_enum(Task, 'status', query)
    query = filter_query_bool(Task, 'sandbox', query)
    return paginated_response(query, lambda items: task_schema.dump(items, many=True))


@task_bp.route('/task/<uuid:task_id>/assignments', methods=['GET'])
def list_task_assignments(task_id):
    task = Task.query.get_or_404(task_id)
    query = sorted_query(Assignment, Assignment.query.filter_by(task=task))
    return paginated_response(query, lambda items: assignment_schema.dump(items, many=True))
