import json

import boto3
from flask import Blueprint, current_app, jsonify, request

from stumpf.decorators import handle_boto_client_errors
from stumpf.middleware import google_auth_middleware
from stumpf.models import CreateWorkerQualificationSchema, Worker, WorkerSchema
from stumpf.mturk_utils import ban_worker, get_worker_qualification_score, \
    is_worker_banned, remove_worker_qualification_score, \
    set_worker_qualification_score, unban_worker
from stumpf.pagination import paginated_response

worker_schema = WorkerSchema()

create_worker_qualification_schema = CreateWorkerQualificationSchema()
worker_bp = Blueprint('worker', __name__)
worker_bp.before_request(google_auth_middleware)


@worker_bp.route('/worker', methods=['GET'])
def get_workers():
    query = Worker.query
    if 'worker_id' in request.args:
        query = query.filter_by(mturk_id=request.args['worker_id'])
    return paginated_response(query, lambda items: worker_schema.dump(items, many=True))


@worker_bp.route('/worker/<worker_id>', methods=['GET'])
def get_worker(worker_id):
    worker = Worker.query.get_or_404(worker_id)
    return jsonify(worker_schema.dump(worker))


@worker_bp.route('/worker/<mturk_id>/ban', methods=['PUT'])
@handle_boto_client_errors
def put_ban_worker(mturk_id):
    if current_app.config['ENV'] in ['development', 'testing']:
        return '', 200
    else:
        session = boto3.session.Session()
        mturk = session.client(
            'mturk', endpoint_url=current_app.config['MTURK_PRODUCTION_ENDPOINT_URL'])
        data = json.loads(request.data)
        if (data['ban']):
            return jsonify(ban_worker(mturk, mturk_id)), 200
        else:
            return jsonify(unban_worker(mturk, mturk_id)), 200


@worker_bp.route('/worker/<mturk_id>/ban', methods=['GET'])
@handle_boto_client_errors
def get_worker_banned(mturk_id):
    if current_app.config['ENV'] in ['development', 'testing']:
        return jsonify({'banned': False}), 200
    else:
        session = boto3.session.Session()
        mturk = session.client(
            'mturk', endpoint_url=current_app.config['MTURK_PRODUCTION_ENDPOINT_URL'])
        result = is_worker_banned(mturk, mturk_id)
        return jsonify({
            'banned': result
        }), 200


@worker_bp.route('/worker/<worker_id>/qualification/<qualification_id>', methods=['PUT'])
@handle_boto_client_errors
def put_worker_qualification(worker_id, qualification_id):
    params = create_worker_qualification_schema.load(request.json or {})
    if current_app.config['ENV'] in ['development', 'testing']:
        return '', 200
    else:
        session = boto3.session.Session()
        mturk = session.client(
            'mturk', endpoint_url=current_app.config['MTURK_PRODUCTION_ENDPOINT_URL'])
        score = params['score']
        notify = params['notify']
        qual = set_worker_qualification_score(mturk, worker_id, qualification_id, score, notify)
        return jsonify(qual), 200


@worker_bp.route('/worker/<worker_id>/qualification/<qualification_id>', methods=['GET'])
@handle_boto_client_errors
def get_worker_qualification(worker_id, qualification_id):
    if current_app.config['ENV'] in ['development', 'testing']:
        return '', 200
    else:
        session = boto3.session.Session()
        mturk = session.client(
            'mturk', endpoint_url=current_app.config['MTURK_PRODUCTION_ENDPOINT_URL'])
        qual = get_worker_qualification_score(mturk, worker_id, qualification_id)
        return jsonify(qual), 200


@worker_bp.route('/worker/<worker_id>/qualification/<qualification_id>', methods=['DELETE'])
@handle_boto_client_errors
def delete_worker_qualification(worker_id, qualification_id):
    params = request.json or {}
    if current_app.config['ENV'] in ['development', 'testing']:
        return '', 200
    else:
        session = boto3.session.Session()
        mturk = session.client(
            'mturk', endpoint_url=current_app.config['MTURK_PRODUCTION_ENDPOINT_URL'])
        reason = params.get('reason', '')
        qual = remove_worker_qualification_score(mturk, worker_id, qualification_id, reason)
        return jsonify(qual), 200
