from datetime import datetime, timedelta
import os

import dotenv
import pytest

from stumpf.celery.utils import make_celery
from stumpf.fakes import fake # noqa E402
from stumpf.models import db, Token # noqa E402
from stumpf.mturk_utils import BAN_LIST_QUALIFICATION_ID # noqa E402


@pytest.fixture(autouse=True)
def mock_load_dotenv(monkeypatch):
    def do_nothing(*args, **kwargs):
        pass

    monkeypatch.setattr(dotenv, 'load_dotenv', do_nothing)


@pytest.fixture(autouse=True)
def mock_celery_app(monkeypatch):
    mock_app = make_celery(entrypoint_config=False)
    mock_app.conf.dburi = os.getenv('SQLALCHEMY_TEST_DATABASE_URI',
                                    'postgresql://postgres@localhost/stumpf_testdb')

    mock_app.conf.broker_url = 'memory://'
    mock_app.conf.result_backend = f'db+{mock_app.conf.dburi}'
    mock_app.conf.database_table_names = {
        'task': 'stumpf_result_backend_taskmeta',
        'group': 'stumpf_result_backend_groupmeta'
    }

    mock_app.conf.timezone = 'America/New_York'
    mock_app.conf.imports += ('stumpf.celery.tasks',)

    # Rate limits must be disabled for testing.
    #
    # Context: This is because the rate limiting functionality is managed
    # with something that approximates threading.Timer() objects. For
    # testing we run the worker in a thread and to ensure tasks have
    # access to Flask's application context we specifically push the
    # context in the worker thread.  Unfortunately if the task is
    # scheduled (by the worker thread) to run in a Timer() thread, the
    # Flask app context isn't passed through to the Timer thread and the
    # task will fail with an "RuntimeError: Working outside of application
    # context." error (assuming it accesses the flask context).
    mock_app.conf.worker_disable_rate_limits = True

    monkeypatch.setattr('stumpf.celery.app.app', mock_app)


@pytest.fixture
def app():
    from stumpf.app import create_app  # noqa E402
    os.environ['SQLALCHEMY_DATABASE_URI'] = os.getenv(
            'SQLALCHEMY_TEST_DATABASE_URI', 'postgresql://postgres@localhost/stumpf_testdb')
    app = create_app({
        'ENV': 'testing'
    })

    with app.app_context():
        db.create_all()

    yield app

    with app.app_context():
        views = [
            'view_token', 'view_mturk_token', 'view_assignment_status_time',
            'view_assignment_detail', 'view_assignment_status_change_detail', 'view_sns_events']
        db.session.execute(f'DROP VIEW IF EXISTS {",".join(views)}')
        db.session.commit()
        db.drop_all()


@pytest.fixture
def auth(app):
    with app.app_context():
        user = fake.stumpf_user()
        token = Token(user=user, expires=datetime.utcnow() + timedelta(hours=1))
        db.session.add_all([user, token])
        db.session.commit()
        yield user, token


@pytest.fixture
def task(app, annotation_app):
    with app.app_context():
        user = fake.stumpf_user()
        db.session.add(user)
        db.session.commit()

        task = fake.stumpf_task(annotation_app.id, user.id)
        db.session.add(task)
        db.session.commit()

        yield task


@pytest.fixture
def assignment(app, task):
    with app.app_context():
        assignment = fake.stumpf_assignment(task.id)
        yield assignment


@pytest.fixture
def annotation_app(app):
    with app.app_context():
        aa = fake.stumpf_annotation_app()
        db.session.add(aa)
        db.session.commit()
        yield aa


@pytest.fixture
def qualification_type(app):
    with app.app_context():
        user = fake.stumpf_user()
        db.session.add(user)
        db.session.commit()

        q = fake.stumpf_qualification_type(user.id)
        db.session.add(q)
        db.session.commit()
        yield q


@pytest.fixture
def banlist(app):
    with app.app_context():
        user = fake.stumpf_user()
        db.session.add(user)
        db.session.commit()

        q = fake.stumpf_qualification_type(user.id, BAN_LIST_QUALIFICATION_ID)
        db.session.add(q)
        db.session.commit()
        yield q


@pytest.fixture
def worker(app):
    with app.app_context():
        w = fake.stumpf_worker()
        db.session.add(w)
        db.session.commit()
        yield w
