from celery.backends.database import SessionManager
from celery.contrib.testing import worker
from mock import patch
import pytest

from stumpf.models import Task


@pytest.fixture
def celery_worker(app, celery_worker_pool, celery_worker_parameters):
    from stumpf.celery.app import app as stumpf_celery_app
    # Note that this must be imported to ensure the stumpf celery app
    # picks up the 'ping' shared task, which is used to verify that
    # the worker process is up and accepting tasks.
    from celery.contrib.testing.tasks import ping  # noqa F401

    # Generate an SQLAlchemy session object that is plugged into the
    # Celery results backend db/tables. This has the added benifit of
    # calling SessionManager.prepare_models(...) which will
    # instantiate the database tables.
    result_backend_session = SessionManager().session_factory(
        dburi=stumpf_celery_app.backend.url,
        short_lived_sessions=stumpf_celery_app.backend.short_lived_sessions,
        **stumpf_celery_app.backend.engine_options)

    # Note that the WorkerController used for testing doesn't include
    # the call to the worker_process_init signal on startup, so we
    # need to make sure the flask app context is pushed and available
    # after the threaded worker starts up. This call is idempotent.
    def _push_test_flask_context_on_worker_start(**kwargs):
        app.app_context().push()

    worker.test_worker_started.connect()(_push_test_flask_context_on_worker_start)

    # Start the worker
    with worker.start_worker(stumpf_celery_app, pool=celery_worker_pool,
                             loglevel='info',
                             **celery_worker_parameters) as w:
        yield w

    # Cleanup the results backend tables
    result_backend_session.execute(
        f'DROP TABLE IF EXISTS {stumpf_celery_app.conf.database_table_names["task"]};')
    result_backend_session.execute(
        f'DROP TABLE IF EXISTS {stumpf_celery_app.conf.database_table_names["group"]};')
    result_backend_session.commit()


def test_create_task_hit(app, auth, annotation_app, celery_worker):

    with patch('stumpf.celery.tasks.create_hit') as create_hit:
        create_hit.return_value = {
            'HIT': {
                'HITId': 'some-hit-id',
                'HITGroupId': 'some-hit-group-id'
            }
        }

        with app.test_client() as client:
            resp = client.post(f'/api/v1/task/{annotation_app.id}/hit', json={
                'title': 'some-title',
                'description': 'some-description',
                'max_assignments': 10,
                'assignment_duration_in_seconds': 60,
                'lifetime_in_seconds': 60,
                'reward': 5,
                'sandbox': True
            }, headers={
                'X-Stumpf-Token': auth[1].token
            })

        assert resp.status_code == 202, resp.json
        resp = celery_worker.app.AsyncResult(resp.json).get(5)

        # Task should return the HIT values returend by
        # create_hit. Note that here these are mocked out.
        assert 'HIT' in resp
        assert resp['HIT']['HITId'] == 'some-hit-id'
        assert resp['HIT']['HITGroupId'] == 'some-hit-group-id'

        task = Task.query.filter_by(hit_id='some-hit-id').first()

        assert task is not None

        assert len(task.status_changes) == 1
        assert task.status == 'Assignable'


@pytest.mark.skip('Celery mocking is broken')
def test_create_batch_hits(app, auth, annotation_app, celery_worker):

    with patch('stumpf.celery.tasks.create_hit') as create_hit:
        create_hit.side_effect = [{
            'HIT': {
                'HITId': 'some-hit-id-1',
                'HITGroupId': 'some-hit-group-id-1'
            }
        }, {
            'HIT': {
                'HITId': 'some-hit-id-2',
                'HITGroupId': 'some-hit-group-id-2'
            }
        }, {
            'HIT': {
                'HITId': 'some-hit-id-3',
                'HITGroupId': 'some-hit-group-id-3'
            }
        }, {
            'HIT': {
                'HITId': 'some-hit-id-4',
                'HITGroupId': 'some-hit-group-id-4'
            }
        }]

        with app.test_client() as client:
            resp = client.post(f'/api/v1/task/batch/hit', json={
                'hits':   [
                    {
                        'app_id': f'{annotation_app.id}',
                        'hit_args': {
                            'sandbox': True,
                            'title': 'test task 1',
                            'description': 'a test task',
                            'keywords': 'test_keyword',
                            'max_assignments': 3,
                            'auto_approval_delay_in_seconds': 50,
                            'assignment_duration_in_seconds': 500,
                            'lifetime_in_seconds': 600,
                            'reward': 15,
                            'qualification_requirements': [],
                            'args': {'some': 'args'}
                        }
                    }, {
                        'app_id': f'{annotation_app.id}',
                        'hit_args': {
                            'sandbox': True,
                            'title': 'test task 2',
                            'description': 'a test task',
                            'keywords': 'test_keyword',
                            'max_assignments': 3,
                            'auto_approval_delay_in_seconds': 50,
                            'assignment_duration_in_seconds': 500,
                            'lifetime_in_seconds': 600,
                            'reward': 15,
                            'qualification_requirements': [],
                            'args': {'some': 'args'}
                        }
                    }, {
                        'app_id': f'{annotation_app.id}',
                        'hit_args': {
                            'sandbox': True,
                            'title': 'test task 3',
                            'description': 'a test task',
                            'keywords': 'test_keyword',
                            'max_assignments': 3,
                            'auto_approval_delay_in_seconds': 50,
                            'assignment_duration_in_seconds': 500,
                            'lifetime_in_seconds': 600,
                            'reward': 15,
                            'qualification_requirements': [],
                            'args': {'some': 'args'}
                        }
                    },
                    {
                        'app_id': f'{annotation_app.id}',
                        'hit_args': {
                            'sandbox': True,
                            'title': 'test task 3',
                            'description': 'a test task',
                            'keywords': 'test_keyword',
                            'max_assignments': 3,
                            'auto_approval_delay_in_seconds': 50,
                            'assignment_duration_in_seconds': 500,
                            'lifetime_in_seconds': 600,
                            'reward': 15,
                            'qualification_requirements': [],
                            'args': {'some': 'args'}
                        }
                    }

                    ]}, headers={
                        'X-Stumpf-Token': auth[1].token
                    })

            assert resp.status_code == 202, resp.json

            # Note that this will raise an exception if any of the tasks failed
            results = celery_worker.app.GroupResult.restore(resp.json).get(5)

            expected_hit_ids = set(['some-hit-id-1', 'some-hit-id-2',
                                    'some-hit-id-3', 'some-hit-id-4'])

            actual_hit_ids = {r['HIT']['HITId'] for r in results}

            assert expected_hit_ids <= actual_hit_ids
            assert actual_hit_ids <= expected_hit_ids

            for _id in list(expected_hit_ids):
                task = Task.query.filter_by(hit_id=_id).first()
                assert task is not None
                assert len(task.status_changes) == 1
                assert task.status == 'Assignable'
