from urllib.parse import parse_qs, urlparse

import pytest

from stumpf.fakes import fake
from stumpf.models import Assignment, db, KwitLease, Task


@pytest.fixture
def kwit(app, annotation_app):
    with app.app_context():
        user = fake.stumpf_user()
        db.session.add(user)
        db.session.commit()

        task = fake.stumpf_task(annotation_app.id, user.id)
        task.type_ = 'KWIT'
        task.status = 'Assignable'
        db.session.add(task)
        db.session.commit()

        yield task


def test_kwit_leasing(app, auth, kwit):
    with app.test_client() as client:
        assert kwit.status == 'Assignable'
        lease = client.post(f'/api/v1/task/kwit/{kwit.id}/lease', headers={
            'X-Stumpf-Token': auth[1].token
        })
        assert lease.status_code == 200, lease.json

        # trying to acquire a second lease will fail
        resp = client.post(f'/api/v1/task/kwit/{kwit.id}/lease', headers={
                'X-Stumpf-Token': auth[1].token
        })
        assert resp.status_code == 400, resp.json


@pytest.fixture
def assignable_kwits(app, annotation_app):
    tasks = []

    with app.app_context():
        user = fake.stumpf_user()
        db.session.add(user)
        db.session.commit()

        for _ in range(5):
            task = fake.stumpf_task(annotation_app.id, user.id)
            task.type_ = 'KWIT'
            task.status = 'Assignable'
            db.session.add(task)
            db.session.commit()
            tasks.append(task)

        yield tasks


def test_kwit_listing(app, auth, task, assignable_kwits):
    task.type_ = 'HIT'
    db.session.add(task)
    db.session.commit()

    assignable_kwits[0].status = 'Unassignable'
    db.session.add_all(assignable_kwits)
    db.session.commit()

    with app.test_client() as client:
        resp = client.get('/api/v1/task?type=KWIT&status=Assignable', headers={
            'X-Stumpf-Token': auth[1].token
        })
        assert resp.status_code == 200, resp.json
        assert len(resp.json) == 4, resp.json


# TODO: This test is skipped now because the celery mocking does not work.
#
# When running as is with no redis instance installed it will fail when trying
# to connect to a result broker.  There are potentially a few things going wrong:
#   1. No worker fixture is created.  Copying the worker fixture from test_hit.py
#      also fails to work.
#   2. The mock_celery_app fixture will work *before* the celery tasks are imported,
#      but the tasks will be registered to the app that was mocked the first time
#      time the fixture was executed.  This likely needs to be changed to a session
#      scoped fixture.
#   3. The mock_celery_app fixture doesn't behave like the included `celery_app`
#      fixture.  There may be some magic going on inside that fixture that the
#      rest of the code relies on.
@pytest.mark.skip('Need to fix worker config')
def test_kwit_end_to_end(app, auth, annotation_app):
    with app.test_client() as client:
        resp = client.post(f'/api/v1/task/{annotation_app.id}/kwit', json={
            'sandbox': True,
            'args': {
                'start': 0,
                'end': 300,
            },
            'title': 'Find starting positions for each of the objects in the video.',
            'description': 'Watch a short video and box the first time an object appears',
            'max_assignments': 1,
        }, headers={
            'X-Stumpf-Token': auth[1].token
        })

        assert resp.status_code == 201, resp.json

        task = Task.query.get(resp.json['id'])
        assert task.status == 'Assignable'

        lease = client.post(f'/api/v1/task/kwit/{task.id}/lease', headers={
            'X-Stumpf-Token': auth[1].token
        })
        assert lease.status_code == 200, lease.json

        assignments = Assignment.query.filter_by(task=task).all()
        assert len(assignments) == 1, assignments

        assert task.status == 'Unassignable'
        assert assignments[0].status == 'Accepted'

        expected_query_args = {
            'kwitId': [str(task.id)],
            'assignmentId': [str(assignments[0].id)],
            'workerId': [str(auth[0].id)],
            'start': ['0'],
            'end': ['300']
        }
        actual_query_args = parse_qs(urlparse(lease.json['url']).query)
        assert actual_query_args == expected_query_args

        db.session.refresh(task)
        resp = client.post('/api/mturk/v1/authenticate/kwit', json={
            'assignment_id': assignments[0].id,
            'app_version': '0123456'
        })

        assert resp.status_code == 201, resp.json
        token = resp.json['token']
        assignment = Assignment.query.get(resp.json['assignment']['id'])

        resp = client.post('/api/mturk/v1/answer', json={
            'answer': {
                'some': 'answer'
            }
        }, headers={
            'X-Stumpf-Mturk-Token': token
        })
        assert resp.status_code == 200, resp.json

        resp = client.post(f'/api/mturk/v1/submit', headers={
            'X-Stumpf-Mturk-Token': token
        })
        assert resp.status_code == 200, resp.json

        db.session.refresh(assignment)
        db.session.refresh(task)

        assert assignment.status == 'Submitted'
        assert [x.status for x in assignment.status_changes] == ['Accepted', 'Submitted']
        assert task.status == 'Reviewable'
        assert [x.status for x in task.status_changes] == ['Assignable', 'Unassignable',
                                                           'Reviewable']
        assert KwitLease.query.count() == 0
