from datetime import datetime, timedelta
import json

from mock import patch
import pytest

from stumpf.fakes import fake
from stumpf.models import Assignment, Task


@pytest.fixture(params=[[('HITReviewable', datetime(2018, 7, 3, 23, 11)),
                         ('HITDisposed',
                          datetime(2018, 7, 3, 23, 11) + timedelta(seconds=5))],
                        [('HITDisposed',
                          datetime(2018, 7, 3, 23, 11) + timedelta(seconds=5)),
                         ('HITReviewable', datetime(2018, 7, 3, 23, 11))]],
                ids=['ordered', 'unordered'])
def hit_messages(request, task):
    messages = []

    for event_type, event_timestamp in request.param:
        messages.append(fake.sns_message(events=[
            fake.sns_task_event(HITId=task.hit_id,
                                EventType=event_type,
                                EventTimestamp=event_timestamp.isoformat(
                                    timespec='seconds') + 'Z')]))

    yield task.hit_id, messages


def test_webhook_task_ordering(app, hit_messages):
    with app.test_client() as client:
        with patch('stumpf.views.webhook.verify_signature') as verify_signature:
            verify_signature.return_value = True

            for message in hit_messages[1]:
                resp = client.post('/webhooks/mturk-notification', data=json.dumps(message))
                assert resp.status_code == 200

        task = Task.query.filter_by(hit_id=hit_messages[0]).first()

        assert len(task.status_changes) == 2

        # Reviewable
        assert task.status_changes[0].occurred == datetime(2018, 7, 3, 23, 11)

        # Disposed
        expected_disposed_time = datetime(2018, 7, 3, 23, 11) + timedelta(seconds=5)
        assert task.status_changes[1].occurred == expected_disposed_time
        assert task.status == 'Disposed'


@pytest.fixture(params=[[('AssignmentAccepted', datetime(2018, 7, 3, 23, 11)),
                         ('AssignmentReturned',
                          datetime(2018, 7, 3, 23, 11) + timedelta(seconds=5))],
                        [('AssignmentReturned',
                          datetime(2018, 7, 3, 23, 11) + timedelta(seconds=5)),
                         ('AssignmentAccepted', datetime(2018, 7, 3, 23, 11))]],
                ids=['ordered', 'unordered'])
def assignment_messages(request, task):
    assignment_id = fake.assignment_id()
    messages = []

    for event_type, event_timestamp in request.param:
        messages.append(fake.sns_message(events=[
            fake.sns_assignment_event(HITId=task.hit_id,
                                      AssignmentId=assignment_id,
                                      EventType=event_type,
                                      EventTimestamp=event_timestamp.isoformat(
                                          timespec='seconds') + 'Z',
                                      WorkerId='ABCDEF1234')]))

    yield messages


def test_webhook_assignment_ordering(app, task, assignment_messages):
    with app.test_client() as client:
        with patch('stumpf.views.webhook.verify_signature') as verify_signature:
            verify_signature.return_value = True

            for message in assignment_messages:
                resp = client.post('/webhooks/mturk-notification', data=json.dumps(message))
                assert resp.status_code == 200

        assignment = Assignment.query.order_by(Assignment.created.desc()).first()

        assert len(assignment.status_changes) == 2

        # accepted
        assert assignment.status_changes[0].occurred == datetime(2018, 7, 3, 23, 11)

        # returned
        expected_returned_time = datetime(2018, 7, 3, 23, 11) + timedelta(seconds=5)
        assert assignment.status_changes[1].occurred == expected_returned_time
        assert assignment.status == 'Returned'


def _authenticate(client, task, assignment_id, app_version):
    return client.post('/api/mturk/v1/authenticate', json={
        'mturk_hit_id': task.hit_id,
        'mturk_assignment_id': assignment_id,
        'app_version': app_version,
        'mturk_worker_id': 'bar'
    })


def _sns_accept(client, task, assignment_id):
    message = fake.sns_message(events=[
        fake.sns_assignment_event(HITId=task.hit_id,
                                  AssignmentId=assignment_id,
                                  EventType='AssignmentAccepted',
                                  EventTimestamp=datetime(2018, 7, 3, 23, 11).isoformat(
                                      timespec='seconds') + 'Z',
                                  WorkerId='bar')])
    return client.post('/webhooks/mturk-notification', data=json.dumps(message))


def test_webhook_assignment_ordering_with_authentication_expected(app, task):
    app_version = fake.sha1()
    assignment_id = fake.assignment_id()

    # Expected behavior
    with app.test_client() as client:
        with patch('stumpf.views.webhook.verify_signature') as verify_signature:
            verify_signature.return_value = True

            _authenticate(client, task, assignment_id, app_version)
            _sns_accept(client, task, assignment_id)

            assignment = Assignment.query.order_by(Assignment.created.desc()).first()
            assert assignment.app_version == app_version


# It's possible we receive an AssignmentAccepted message from SNS
# before the application authenticates. In this case we won't have the
# app hash, which only comes from hitting the authenticate endpoint.
# Make sure the code correctly updates the hash once the authenticate
# endpoint is hit.
def test_webhook_assignment_ordering_with_authentication_edgecase(app, task):
    app_version = fake.sha1()
    assignment_id = fake.assignment_id()

    # Edge case
    with app.test_client() as client:
        with patch('stumpf.views.webhook.verify_signature') as verify_signature:
            verify_signature.return_value = True

            _sns_accept(client, task, assignment_id)
            _authenticate(client, task, assignment_id, app_version)

            assignment = Assignment.query.order_by(Assignment.created.desc()).first()
            assert assignment.app_version == app_version


@pytest.mark.skip(reason='Signature verification is temporarily disabled (see 6dfbe0b).')
def test_webhook_authorization(app):
    with app.test_client() as client:
        resp = client.post('/webhooks/mturk-notification', data=json.dumps({
            'SigningCertURL': 'https://some-scam-website.com',
            'Message': json.dumps({})
        }))

    assert resp.status_code == 403

    with app.test_client() as client:
        resp = client.post('/webhooks/mturk-notification', data=json.dumps({
            'SigningCertURL':
            'https://sns.us-east-2.amazonaws.com/SimpleNotificationService-12345.pem',
            'Message': json.dumps({}),
            'Signature': 'some-bad-signature'
        }))

    assert resp.status_code == 401
